"""
Constructing custom probability distributions is done in one of two ways:
Sublcassing the :class:`~chaospy.distributions.Dist` or by calling
:func:`~chaospy.distributions.construct`. They work about the same except for one
methods are defined, while in the other, functions.

Import the construction function::

    >>> from chaospy.distributions import construct, Dist

A simple example for constructing a simple uniform distribution::

    >>> def cdf(self, x, lo, up):
    ...     return (x-lo)/(up-lo)
    >>> def bnd(self, lo, up):
    ...     return lo, up
    >>> Uniform = construct(cdf=cdf, bnd=bnd)
    >>> dist = Uniform(lo=-3, up=3)
    >>> print(dist.fwd([-3, 0, 3]))
    [0.  0.5 1. ]

Here ``cdf`` is the dependent cumulative distribution function as defined in
equation , ``bnd`` is a function returning the lower and upper bounds, and
``a`` and ``b`` are distribution parameters.  They take either other
components, or as illustrated: constants.


In addition to ``cdf`` and ``bnd`` there are a few optional arguments. For
example a fully featured uniform random variable is defined as follows::

    >>> def pdf(self, x, lo, up):
    ...     return 1./(up-lo)
    >>> def ppf(self, q, lo, up):
    ...     return q*(up-lo) + lo
    >>> Uniform = construct(
    ...     cdf=cdf, bnd=bnd, pdf=pdf, ppf=ppf)
    >>> dist = Uniform(lo=-3, up=3)

There ``pdf`` is probability distribution function and ``ppf`` if the point
percentile function.  These are methods that provides needed functionality for
probabilistic collocation. If they are not provided during construct, they are
estimated as far as possible.

Equivalently constructing the same distribution using subclass:
:func:`~chaospy.distributions.construct`::

    >>> class Uniform(Dist):
    ...     def __init__(self, lo=0, up=1):
    ...         Dist.__init__(self, lo=lo, up=up)
    ...     def _cdf(self, x, lo, up):
    ...         return (x-lo)/(up-lo)
    ...     def _bnd(self, lo, up):
    ...         return lo, up
    ...     def _pdf(self, x, lo, up):
    ...         return 1./(up-lo)
    ...     def _ppf(self, q, lo, up):
    ...         return q*(up-lo) + lo
    ...     def _str(self, lo, up):
    ...         return "u(%s%s)" % (lo, up)
    >>> dist = Uniform(-3, 3)
    >>> print(dist.fwd([-3, 0, 3])) # Forward Rosenblatt transformation
    [0.  0.5 1. ]

"""
import types
import numpy as np


class Dist(object):
    """
    The distribution backend class.

    Subclass this module to construct a custom distribution.

    If direct subclass of Dist, two method must be provided:

    * Cumulative distribution function (CDF): ``_cdf(self, x, **prm)``.
    * Upper and lower bounds ``_bnd(self, **prm)``.

    The following can be provided:

    * Probability density function: ``_pdf(self, x, **prm)``.
    * CDF inverse: ``_ppf(self, q, **prm)``.
    * Statistical moment generator: ``_mom(self, k, **prm)``.
    * TTR coefficients generator: ``_ttr(self, k, **prm)``.
    * Pretty print of distribution: ``_str(self, **prm)``.

    Alternative use the construct generator
    :func:`~chaospy.distributions.construct`.
    """

    __array_priority__ = 9000

    def __init__(self, **prm):
        """
        Args:
            _length (int) : Length of the distribution
            _advanced (bool) : If True, activate advanced mode
            **prm (array_like) : Other optional parameters. Will be assumed when
                    calling any sub-functions.
        """
        from . import graph
        for key, val in prm.items():
            if not isinstance(val, Dist):
                prm[key] = np.array(val)

        self.length = int(prm.pop("_length", 1))
        self.advance = prm.pop("_advance", False)
        self.prm = prm.copy()
        self.graph = graph.Graph(self)
        self.dependencies = self.graph.run(self.length, "dep")[0]

    def range(self, x=None, retall=False, verbose=False):
        """
        Generate the upper and lower bounds of a distribution.

        Args:
            x (array_like, optional) : The bounds might vary over the sample
                    space. By providing x you can specify where in the space
                    the bound should be taken.  If omited, a (pseudo-)random
                    sample is used.

        Returns:
            (np.ndarray) : The lower (out[0]) and upper (out[1]) bound where
                    out.shape=(2,)+x.shape
        """
        dim = len(self)
        if x is None:
            from . import approx
            x = approx.find_interior_point(self)
        else:
            x = np.array(x)
        shape = x.shape
        size = int(x.size/dim)
        x = x.reshape(dim, size)

        out, graph = self.graph.run(x, "range")
        out = out.reshape((2,)+shape)

        if verbose>1:
            print(graph)

        if retall:
            return out, graph
        return out

    def fwd(self, x):
        """
        Forward Rosenblatt transformation.

        Args:
            x (array_like) : Location for the distribution function. x.shape
                    must be compatible with distribution shape.

        Returns:
            (ndarray) : Evaluated distribution function values, where
                    out.shape==x.shape.
        """
        from . import rosenblatt
        return rosenblatt.fwd(self, x)

    def cdf(self, x):
        """
        Cumulative distribution function.

        Note that chaospy only supports cumulative distribution funcitons in
        one dimensions.

        Args:
            x (array_like) : Location for the distribution function. x.shape
                    must be compatible with distribution shape.

        Returns:
            (ndarray) : Evaluated distribution function values, where
                    out.shape==x.shape.

        Except:
            (NotImplementedError) : for distributions with dependent
                components.
        """
        if self.dependent():
            raise NotImplementedError("""\
Cumulative distribution function is only available for stocastically \
independent variables""")
        from . import rosenblatt
        out = rosenblatt.fwd(self, x)
        if len(self) > 1:
            out = np.prod(out, 0)
        return out


    def inv(self, q, maxiter=100, tol=1e-5, verbose=False, **kws):
        """
        Inverse Rosenblatt transformation.

        Args:
            q (array_like) : Probabilities to be inverse. If any values are
                    outside [0,1], error will be raised. q.shape must be
                    compatible with diistribution shape.

        Kwargs:
            maxiter (int) : Maximum number of iterations
            tol (float) : Tolerence level

        Returns:
            (ndarray) : Inverted probability values where out.shape==q.shape.
        """
        from . import rosenblatt
        return rosenblatt.inv(self, q, maxiter, tol, **kws)

    def pdf(self, x, step=1e-7, verbose=0):
        """
        Probability density function.

        Args:
            x (array_like) : Location for the density function. x.shape must
                    be compatible with distribution shape.
            step (float, array_like) : The step length given aproximation is
                    used. If array provided, elements are used along each
                    axis.

        Returns:
            (ndarray) : Evaluated density function values. Shapes are related
                    through the identity x.shape=dist.shape+out.shape
        """
        dim = len(self)
        x = np.array(x)
        shape = x.shape
        size = int(x.size/dim)
        x = x.reshape(dim, size)
        out = np.zeros((dim, size))

        (lo, up), graph = self.graph.run(x, "range")
        valids = np.prod((x.T >= lo.T)*(x.T <= up.T), 1, dtype=bool)
        x[:, ~valids] = (.5*(up+lo))[:, ~valids]
        out = np.zeros((dim,size))

        try:
            tmp,graph = self.graph.run(x, "pdf",
                    eps=step)
            out[:,valids] = tmp[:,valids]
        except NotImplementedError:
            from . import approx
            tmp, graph = approx.pdf_full(self, x, step, retall=True)
            out[:, valids] = tmp[:, valids]
            if verbose:
                print("approx %s.pdf")
        except IndexError:
            pass

        if verbose>1:
            print(self.graph)

        out = out.reshape(shape)
        if dim>1:
            out = np.prod(out, 0)
        return out

    def sample(self, size=(), rule="R", antithetic=None,
            verbose=False, **kws):
        """
        Create pseudo-random generated samples.

        Args:
            size (int,array_like):
                The size of the samples to generate.
            rule (str):
                Alternative sampling techniques. See
                :func:`~chaospy.distributions.sampler.generator.generate_samples`.
            antithetic (bool, array_like):
                If provided, will be used to setup antithetic variables. If
                array, defines the axes to mirror.

        Returns:
            (ndarray) : Random samples with shape (len(self),)+self.shape
        """
        size_ = np.prod(size, dtype=int)
        dim = len(self)
        if dim > 1:
            if isinstance(size, (tuple,list,np.ndarray)):
                shape = (dim,) + tuple(size)
            else:
                shape = (dim, size)
        else:
            shape = size

        from . import sampler
        out = sampler.generator.generate_samples(
            order=size_, domain=self, rule=rule, antithetic=antithetic)
        try:
            out = out.reshape(shape)
        except:
            if len(self) == 1:
                out = out.flatten()
            else:
                out = out.reshape(dim, int(out.size/dim))

        return out

    def mom(self, K, **kws):
        """
        Raw statistical moments.

        Creates non-centralized raw moments from the random variable. If
        analytical options can not be utilized, Monte Carlo integration
        will be used.

        Args:
            K (array_like) : Index of the raw moments. k.shape must be
                    compatible with distribution shape.  Sampling scheme when
                    performing Monte Carlo
            rule (str) : rule for estimating the moment if the analytical
                    method fails.
            composit (int, array_like optional) : If provided, composit
                    quadrature will be used.  Ignored in the case if
                    gaussian=True.  If int provided, determines number of even
                    domain splits. If array of ints, determines number of even
                    domain splits along each axis. If array of arrays/floats,
                    determines location of splits.
            antithetic (array_like, optional) : List of bool. Represents the
                    axes to mirror using antithetic variable during MCI.

        Returns:
            (ndarray) : Shapes are related through the identity
                    `k.shape==dist.shape+k.shape`.
        """
        K = np.array(K, dtype=int)
        shape = K.shape
        dim = len(self)

        if dim > 1:
            shape = shape[1:]

        size = int(K.size/dim)
        K = K.reshape(dim, size)

        try:
            out, _ = self.graph.run(K, "mom", **kws)
        except NotImplementedError:
            from . import approx
            out = approx.mom(self, K, **kws)

        return out.reshape(shape)

    def ttr(self, k, acc=10**3, verbose=1):
        """
        Three terms relation's coefficient generator

        Args:
            k (array_like, int) : The order of the coefficients.
            acc (int) : Accuracy of discretized Stieltjes if analytical
                    methods are unavailable.

        Returns:
            (Recurrence coefficients) : Where out[0] is the first (A) and
                    out[1] is the second coefficient With
                    `out.shape==(2,)+k.shape`.
        """
        k = np.array(k, dtype=int)
        dim = len(self)
        shape = k.shape
        shape = (2,) + shape
        size = int(k.size/dim)
        k = k.reshape(dim, size)

        out, graph = self.graph.run(k, "ttr")
        return out.reshape(shape)

    def _ttr(self, *args, **kws):
        """Default TTR generator, throws error."""
        raise NotImplementedError

    def _dep(self, graph):
        """
        Default dependency module backend.

        See graph for advanced distributions.
        """
        sets = [graph(dist) for dist in graph.dists]
        if len(self)==1:
            out = [set([self])]
        else:
            out = [set([]) for _ in range(len(self))]
        for set_ in sets:
            for idx in range(len(self)):
                out[idx].update(set_[idx])
        return out


    def __str__(self):
        """X.__str__() <==> str(X)"""
        if hasattr(self, "_str"):
            return str(self._str(**self.prm))
        return "D"

    def __len__(self):
        """X.__len__() <==> len(X)"""
        return self.length

    def __add__(self, X):
        """Y.__add__(X) <==> X+Y"""
        from . import operators
        return operators.add(self, X)

    def __radd__(self, X):
        """Y.__radd__(X) <==> Y+X"""
        from . import operators
        return operators.add(self, X)

    def __sub__(self, X):
        """Y.__sub__(X) <==> X-Y"""
        from . import operators
        return operators.add(self, -X)

    def __rsub__(self, X):
        """Y.__rsub__(X) <==> Y-X"""
        from . import operators
        return operators.add(X, -self)

    def __neg__(self):
        """X.__neg__() <==> -X"""
        from . import operators
        return operators.neg(self)

    def __mul__(self, X):
        """Y.__mul__(X) <==> X*Y"""
        from . import operators
        return operators.mul(self, X)

    def __rmul__(self, X):
        """Y.__rmul__(X) <==> Y*X"""
        from . import operators
        return operators.mul(self, X)

    def __div__(self, X):
        """Y.__div__(X) <==> Y/X"""
        from . import operators
        return operators.mul(self, X**-1)

    def __rdiv__(self, X):
        """Y.__rdiv__(X) <==> X/Y"""
        from . import operators
        return operators.mul(X, self**-1)

    def __floordiv__(self, X):
        """Y.__floordiv__(X) <==> Y/X"""
        from . import operators
        return operators.mul(self, X**-1)

    def __rfloordiv__(self, X):
        """Y.__rfloordiv__(X) <==> X/Y"""
        from . import operators
        return operators.mul(X, self**-1)

    def __truediv__(self, X):
        """Y.__truediv__(X) <==> Y/X"""
        from . import operators
        return operators.mul(self, X**-1)

    def __rtruediv__(self, X):
        """Y.__rtruediv__(X) <==> X/Y"""
        from . import operators
        return operators.mul(X, self**-1)

    def __pow__(self, X):
        """Y.__pow__(X) <==> Y**X"""
        from . import operators
        return operators.pow(self, X)

    def __rpow__(self, X):
        """Y.__rpow__(X) <==> X**Y"""
        from . import operators
        return operators.pow(X, self)

    def __le__(self, X):
        """Y.__le__(X) <==> Y<=X"""
        from . import operators
        return operators.trunk(self, X)

    def __lt__(self, X):
        """Y.__lt__(X) <==> Y<X"""
        from . import operators
        return operators.trunk(self, X)

    def __ge__(self, X):
        """Y.__ge__(X) <==> Y>=X"""
        from . import operators
        return operators.trunk(X, self)

    def __gt__(self, X):
        """Y.__gt__(X) <==> Y>X"""
        from . import operators
        return operators.trunk(X, self)

    def addattr(self, **kws):
        """
        Add attribution to distribution

        Kwargs:
            pdf (callable) : Probability density function.
            cdf (callable) : Cumulative distribution function.
            ppf (callable) : Point percentile function.
            mom (callable) : Raw statistical moments.
            ttr (callable) : Three term recursion coefficient generator.
            val (callable) : If auxiliary variable, try to return the values
                    of it's underlying variables, else return self.
            str (callable, str) : Pretty print of module.
            dep (callable) : Dependency structure (if non-trivial).
        """
        for key,val in kws.items():
            if key == "str" and isinstance(val, str):
                val_ = val
                val = lambda *a,**k: val_
            setattr(self, "_"+key, types.MethodType(val, self))


    def dependent(self, *args):
        """
        Determine dependency structure in module

        Args:
            arg, [...] (optional) : If omited, internal dependency will be
                    determined.  If included, dependency between self and args
                    will be used.

        Returns:
            (bool) : True if distribution is dependent.
        """
        sets, graph = self.graph.run(None, "dep")

        if args:

            sets_ = set()
            for set_ in sets:
                sets_ = sets_.union(set_)
            sets = [sets_]

            for arg in args:
                sets_ = set()
                for set_ in arg.graph.run(None, "dep"):
                    sets_ = sets_.union(set_)
                sets.append(sets_)

        as_seperated = sum([len(set_) for set_ in sets])

        as_joined = set()
        for set_ in sets:
            as_joined = as_joined.union(set_)
        as_joined = len(as_joined)

        return as_seperated != as_joined



r"""
Creating an orthogonal polynomial expansion can be numerical unstable
when using raw statistical moments as input
:cite:`gautschi_construction_1968`. This can be a problem if
constructing large expansions since the error blows up. Given that the
distribution is univariate it is instead possible to create orthogonal
polynomials stabilized using the three terms recursion relation:

.. math::

    \Phi_{n+1}(q) = \Phi_{n}(q)(q-A_n) - \Phi_{n-1}(q) B_n,

where

.. math::

    A_n &= \frac{
        \left\langle
            q \Phi_n, \Phi_n
        \right\rangle
    }{
        \left\langle
            \Phi_n, \Phi_n
        \right\rangle
    } = \frac{
        \mathbb E[q\Phi_{n}^2]
    }{
        \mathbb E[\Phi_n^2]
    }

    B_n &= \frac{
        \left\langle
            \Phi_n, \Phi_n
        \right\rangle
    }{
        \left\langle
            \Phi_{n-1}, \Phi_{n-1}
        \right\rangle
    } = \frac{
        \mathbb E[\Phi_{n}^2]
    }{
        \mathbb E[\Phi_{n-1}^2]
    }

A multivariate polynomial expansion can be created using tensor product
rule of univariate polynomials expansions. This assumes that the
distribution is stochastically independent.

In the ``chaospy`` toolbox three terms recursion coefficient can be
generating by calling the ``ttr`` instance method::

    >>> dist = chaospy.Uniform(-1,1)
    >>> print(numpy.around(dist.ttr([0,1,2,3]), 4))
    [[ 0.      0.      0.      0.    ]
     [-0.      0.3333  0.2667  0.2571]]

Looking back to section :ref:`distributions` and the ``pc.contruct`` function,
``ttr`` can be added as a keyword argument. So tailored recursion coefficients
can be added for user defined distributions. If the ``ttr`` function is
missing, which can often the case, the coefficients can be calculated using
discretized Stieltjes method :cite:`gautschi_construction_1968`. The method
consists of alternating between calculating expression and expression given
starting values :math:`\Phi_{-1}=0` and :math:`\Phi_{0}=1`. Since the expected
value operator is used, this method can also be considered as a statistical
moment based method, however the coefficients calculations in equation can be
estimated using numerical integration, and made stable. In ``chaospy`` if the
``ttr`` is missing, it is estimated using ``quadgen`` with Clenshaw-Curtis
nodes and weights. The default is order 40, however, as with all the other
instance methods so far, it is
possible to set the wanted parameters using keyword argument. In this
case the keyword argument ``acc`` can be used to change the default. In
section :ref:`moments` the ``moment_generator`` function was introduced.
Analogous there is also a ``recurrence_generator`` function that does the same,
but for the ``ttr``. In other words, it is possible to fix the parameters in
the estimation of ``ttr`` in any distribution. Note that the keyword
``rule="G"`` is disabled since the Golub-Welsch algorithm also depends
upon the three terms recursion coefficients for it's calculations
:cite:`golub_calculation_1967`.

Multivariate orthogonal polynomial expansions are created by multiplying
univariate polynomials together:

.. math::

    \Phi_n = \Phi_{1,n_1}\cdots\Phi_{N,n_N}

where :math:`\Phi_{i,n_i}` represents the :math:`n_i`-th polynomial in
the univariate expansion orthogonal with respect to the :math:`i`-th
component of :math:`Q`. For the orthogonality to hold, it must be
assumed that :math:`p_{Q}` is stochastically independent. This to
assure the third equality in

.. math::

    \left\langle \Phi_n, \Phi_m \right\rangle &=
    \mathbb E[\Phi_n \Phi_m ] =
    \mathbb E[ \Phi_{1,n_1}\Phi_{i,m_1} \cdots\Phi_{N,n_N}
    \Phi_{N,m_N} ]

    &= \mathbb E[\Phi_{1,n_1}\Phi_{1,m_1}]\cdots
    \mathbb E[ \Phi_{N,n_N}\Phi_{N,m_N} ]

    &= \left\langle{\Phi_{1,n_1},\Phi_{1,m_1}}\right\rangle
    \cdots \left\langle{\Phi_{N,n_N},\Phi_{N,m_N}}\right\rangle.

Since each univariate polynomial expansion is orthogonal, this implies
that the multivariate also is orthogonal.

In ``chaospy`` constructing orthogonal polynomial using the three term
recursion scheme can be done through ``orth_ttr``. For example::

    >>> dist = chaospy.Iid(chaospy.Gamma(1), 2)
    >>> orths = chaospy.orth_ttr(2, dist)
    >>> print(chaospy.around(orths, 4))
    [1.0, q1-1.0, q0-1.0, q1^2-4.0q1+2.0, q0q1-q0-q1+1.0, q0^2-4.0q0+2.0]

The method will use the ``ttr`` function if available, and discretized
Stieltjes otherwise.
"""
import logging

import numpy
import chaospy

__all__ = (
    "orth_gs",
    "orth_ttr",
    "orth_chol",
    "norm",
)


def orth_gs(order, dist, normed=False, sort="GR", cross_truncation=1., **kws):
    """
    Gram-Schmidt process for generating orthogonal polynomials.

    Args:
        order (int, Poly):
            The upper polynomial order. Alternative a custom polynomial basis
            can be used.
        dist (Dist):
            Weighting distribution(s) defining orthogonality.
        normed (bool):
            If True orthonormal polynomials will be used instead of monic.
        sort (str):
            Ordering argument passed to poly.basis. If custom basis is used,
            argument is ignored.
        cross_truncation (float):
            Use hyperbolic cross truncation scheme to reduce the number of
            terms in expansion.
        kws (optional):
            Keyword argument passed to dist.mom if the moments need to be
            estimated.

    Returns:
        (Poly) : The orthogonal polynomial expansion.

    Examples:
        >>> Z = chaospy.J(chaospy.Normal(), chaospy.Normal())
        >>> print(chaospy.around(chaospy.orth_gs(2, Z), 4))
        [1.0, q1, q0, q1^2-1.0, q0q1, q0^2-1.0]
    """
    logger = logging.getLogger(__name__)
    dim = len(dist)

    if isinstance(order, int):
        if order == 0:
            return chaospy.poly.Poly(1, dim=dim)
        basis = chaospy.poly.basis(
            0, order, dim, sort, cross_truncation=cross_truncation)
    else:
        basis = order

    basis = list(basis)

    polynomials = [basis[0]]

    if normed:
        for idx in range(1, len(basis)):

            # orthogonalize polynomial:
            for idy in range(idx):
                orth = chaospy.descriptives.E(
                    basis[idx]*polynomials[idy], dist, **kws)
                basis[idx] = basis[idx] - polynomials[idy]*orth

            # normalize:
            norms = chaospy.descriptives.E(polynomials[-1]**2, dist, **kws)
            if norms <= 0:
                logger.warning("Warning: Polynomial cutoff at term %d", idx)
                break
            basis[idx] = basis[idx] / numpy.sqrt(norms)

            polynomials.append(basis[idx])

    else:

        norms = [1.]
        for idx in range(1, len(basis)):

            # orthogonalize polynomial:
            for idy in range(idx):
                orth = chaospy.descriptives.E(
                    basis[idx]*polynomials[idy], dist, **kws)
                basis[idx] = basis[idx] - polynomials[idy] * orth / norms[idy]

            norms.append(
                chaospy.descriptives.E(polynomials[-1]**2, dist, **kws))
            if norms[-1] <= 0:
                logger.warning("Warning: Polynomial cutoff at term %d", idx)
                break

            polynomials.append(basis[idx])

    return chaospy.poly.Poly(polynomials, dim=dim, shape=(len(polynomials),))


def orth_ttr(
        order, dist, normed=False, sort="GR", retall=False,
        cross_truncation=1., **kws):
    """
Create orthogonal polynomial expansion from three terms recursion formula.

Args:
    order (int):
        Order of polynomial expansion.
    dist (Dist):
        Distribution space where polynomials are orthogonal If dist.ttr
        exists, it will be used. Must be stochastically independent.
    normed (bool):
        If True orthonormal polynomials will be used.
    sort (str):
        Polynomial sorting. Same as in basis.
    retall (bool):
        If true return numerical stabalized norms as well. Roughly the same as
        ``cp.E(orth**2, dist)``.
    cross_truncation (float):
        Use hyperbolic cross truncation scheme to reduce the number of
        terms in expansion. only include terms where the exponents ``K``
        satisified the equation
        ``order >= sum(K**(1/cross_truncation))**cross_truncation``.
    **kws (optional):
        Keyword argument passed to Stieltjes function ``generate_stieltjes``.

Returns:
    orth (Poly, numpy.ndarray):
        Orthogonal polynomial expansion and norms of the orthogonal
        expansion on the form E(orth**2, dist). Calculated using recurrence
        coefficients for stability.

Examples:
    >>> Z = chaospy.Normal()
    >>> print(chaospy.around(chaospy.orth_ttr(4, Z), 4))
    [1.0, q0, q0^2-1.0, q0^3-3.0q0, q0^4-6.0q0^2+3.0]
    """
    polynomials, norms, _, _ = chaospy.quad.generate_stieltjes(
        dist=dist, order=order, retall=True, **kws)

    if normed:
        for idx, poly in enumerate(polynomials):
            polynomials[idx] = poly / numpy.sqrt(norms[:, idx])
        norms = norms**0

    dim = len(dist)
    if dim > 1:
        mv_polynomials = []
        mv_norms = []
        indices = chaospy.bertran.bindex(
            start=0, stop=order, dim=dim, sort=sort,
            cross_truncation=cross_truncation,
        )

        for index in indices:
            poly = polynomials[index[0]][0]
            for idx in range(1, dim):
                poly = poly * polynomials[index[idx]][idx]
            mv_polynomials.append(poly)

        if retall:
            for index in indices:
                mv_norms.append(
                    numpy.prod([norms[idx, index[idx]] for idx in range(dim)]))

    else:
        mv_norms = norms[0]
        mv_polynomials = polynomials

    polynomials = chaospy.poly.flatten(chaospy.poly.Poly(mv_polynomials))

    if retall:
        return polynomials, numpy.array(mv_norms)
    return polynomials


def orth_chol(order, dist, normed=True, sort="GR", cross_truncation=1., **kws):
    """
    Create orthogonal polynomial expansion from Cholesky decomposition.

    Args:
        order (int) : Order of polynomial expansion
        dist (Dist) : Distribution space where polynomials are orthogonal
        normed (bool) : If True orthonormal polynomials will be used instead
                of monic.
        sort (str) : Ordering argument passed to poly.basis.  If custom basis
                is used, argument is ignored.
        cross_truncation (float) : Use hyperbolic cross truncation scheme to
                reduce the number of terms in expansion.
        kws (optional) : Keyword argument passed to dist.mom.

    Examples:
        >>> Z = chaospy.Normal()
        >>> print(chaospy.around(chaospy.orth_chol(3, Z), 4))
        [1.0, q0, 0.7071q0^2-0.7071, 0.4082q0^3-1.2247q0]
    """
    dim = len(dist)
    basis = chaospy.poly.basis(
        start=1, stop=order, dim=dim, sort=sort,
        cross_truncation=cross_truncation,
    )
    length = len(basis)

    cholmat = chaospy.chol.gill_king(chaospy.descriptives.Cov(basis, dist))
    cholmat_inv = numpy.linalg.inv(cholmat.T).T
    if not normed:
        diag_mesh = numpy.repeat(numpy.diag(cholmat_inv), len(cholmat_inv))
        cholmat_inv /= diag_mesh.reshape(cholmat_inv.shape)

    coefs = numpy.empty((length+1, length+1))

    coefs[1:, 1:] = cholmat_inv
    coefs[0, 0] = 1
    coefs[0, 1:] = 0

    expected = -numpy.sum(
        cholmat_inv*chaospy.descriptives.E(basis, dist, **kws), -1)
    coefs[1:, 0] = expected

    coefs = coefs.T

    out = {}
    out[(0,)*dim] = coefs[0]
    for idx in range(length):
        index = basis[idx].keys[0]
        out[index] = coefs[idx+1]

    polynomials = chaospy.poly.Poly(out, dim, coefs.shape[1:], float)

    return polynomials



def norm(order, dist, orth=None):

    dim = len(dist)
    try:
        if dim>1:
            norms = numpy.array([norm(order+1, D) for D in dist])
            Is = chaospy.bertran.bindex(order, dim)
            out = numpy.ones(len(Is))

            for i in range(len(Is)):
                index = Is[i]
                for j in range(dim):
                    if index[j]:
                        out[i] *= norms[j, index[j]]
            return out

        K = range(1,order+1)
        ttr = [1.] + [dist.ttr(k)[1] for k in K]
        return numpy.cumprod(ttr)

    except NotImplementedError:

        if orth is None:
            orth = orth_chol(order, dist)
        return chaospy.descriptives.E(orth**2, dist)


def lagrange_polynomial(absicas, sort="GR"):
    """
    Create Lagrange polynomials.

    Args:
        absicas (array_like):
            Sample points where the Lagrange polynomials shall be defined.

    Example:
        >>> print(chaospy.around(lagrange_polynomial([-10, 10]), 4))
        [-0.05q0+0.5, 0.05q0+0.5]
        >>> print(chaospy.around(lagrange_polynomial([-1, 0, 1]), 4))
        [0.5q0^2-0.5q0, -q0^2+1.0, 0.5q0^2+0.5q0]
        >>> poly = lagrange_polynomial([[1, 0, 1], [0, 1, 2]])
        >>> print(chaospy.around(poly, 4))
        [0.5q0-0.5q1+0.5, -q0+1.0, 0.5q0+0.5q1-0.5]
        >>> print(numpy.around(poly([1, 0, 1], [0, 1, 2]), 4))
        [[1. 0. 0.]
         [0. 1. 0.]
         [0. 0. 1.]]
    """
    absicas = numpy.asfarray(absicas)
    if len(absicas.shape) == 1:
        absicas = absicas.reshape(1, absicas.size)
    dim, size = absicas.shape

    order = 1
    while chaospy.bertran.terms(order, dim) <= size:
        order += 1

    indices = numpy.array(chaospy.bertran.bindex(0, order-1, dim, sort)[:size])
    idx, idy = numpy.mgrid[:size, :size]

    matrix = numpy.prod(absicas.T[idx]**indices[idy], -1)
    det = numpy.linalg.det(matrix)
    if det == 0:
        raise numpy.linalg.LinAlgError("invertible matrix")

    vec = chaospy.poly.basis(0, order-1, dim, sort)[:size]

    coeffs = numpy.zeros((size, size))

    if size == 1:
        out = chaospy.poly.basis(0, 0, dim, sort)*absicas.item()

    elif size == 2:
        coeffs = numpy.linalg.inv(matrix)
        out = chaospy.poly.sum(vec*(coeffs.T), 1)

    else:
        for i in range(size):
            for j in range(size):
                coeffs[i, j] += numpy.linalg.det(matrix[1:, 1:])
                matrix = numpy.roll(matrix, -1, axis=0)
            matrix = numpy.roll(matrix, -1, axis=1)
        coeffs /= det
        out = chaospy.poly.sum(vec*(coeffs.T), 1)

    return out



"""
Implementation of Stieltjes' method.
"""
import numpy

import chaospy.poly.base
import chaospy.poly.collection
import chaospy.quad


def generate_stieltjes(
        dist, order, accuracy=100, normed=False, retall=False, **kws):
    """
    Discretized Stieltjes' method.

    Args:
        dist (Dist):
            Distribution defining the space to create weights for.
        order (int):
            The polynomial order create.
        accuracy (int):
            The quadrature order of the Clenshaw-Curtis nodes to
            use at each step, if approximation is used.
        retall (bool):
            If included, more values are returned

    Returns:
        (list): List of polynomials, norms of polynomials and three terms
            coefficients. The list created from the method with
            ``len(orth) == order+1``. If ``len(dist) > 1``, then each
            polynomials are multivariate.
        (numpy.ndarray, numpy.ndarray, numpy.ndarray): If ``retall`` is true,
            also return polynomial norms and the three term coefficients.
            The norms of the polynomials with ``norms.shape = (dim, order+1)``
            where ``dim`` are the number of dimensions in dist.  The
            coefficients have ``shape == (dim, order+1)``.

    Examples:
        >>> dist = chaospy.J(chaospy.Normal(), chaospy.Weibull())
        >>> orth, norms, coeffs1, coeffs2 = chaospy.generate_stieltjes(
        ...     dist, 2, retall=True)
        >>> print(chaospy.around(orth[2], 5))
        [q0^2-1.0, q1^2-4.0q1+2.0]
        >>> print(numpy.around(norms, 5))
        [[1. 1. 2.]
         [1. 1. 4.]]
        >>> print(numpy.around(coeffs1, 5))
        [[0. 0. 0.]
         [1. 3. 5.]]
        >>> print(numpy.around(coeffs2, 5))
        [[1. 1. 2.]
         [1. 1. 4.]]

        >>> dist = chaospy.Uniform()
        >>> orth, norms, coeffs1, coeffs2 = chaospy.generate_stieltjes(
        ...     dist, 2, retall=True)
        >>> print(chaospy.around(orth[2], 8))
        q0^2-q0+0.16666667
        >>> print(numpy.around(norms, 4))
        [[1.     0.0833 0.0056]]
    """
    assert not dist.dependent()

    if len(dist) > 1:

        # one for each dimension:
        orth, norms, coeff1, coeff2 = zip(*[generate_stieltjes(
            _, order, accuracy, normed, retall=True, **kws) for _ in dist])

        # ensure each polynomial has its own dimension:
        orth = [[chaospy.setdim(_, len(orth)) for _ in poly] for poly in orth]
        orth = [[chaospy.rolldim(_, len(dist)-idx) for _ in poly] for idx, poly in enumerate(orth)]
        orth = [chaospy.poly.base.Poly(_) for _ in zip(*orth)]

        if not retall:
            return orth

        # stack results:
        norms = numpy.vstack(norms)
        coeff1 = numpy.vstack(coeff1)
        coeff2 = numpy.vstack(coeff2)

        return orth, norms, coeff1, coeff2

    try:
        orth, norms, coeff1, coeff2 = _stieltjes_analytical(
            dist, order, normed)
    except NotImplementedError:
        orth, norms, coeff1, coeff2 = _stieltjes_approx(
            dist, order, accuracy, normed, **kws)

    if retall:
        return orth, norms, coeff1, coeff2
    return orth


def _stieltjes_analytical(dist, order, normed):
    """Stieltjes' method with analytical recurrence coefficients."""
    dimensions = len(dist)
    mom_order = numpy.arange(
        order+1
    ).repeat(
        dimensions
    ).reshape(
        order+1, dimensions
    ).T
    coeff1, coeff2 = dist.ttr(mom_order)
    coeff2[:, 0] = 1.

    poly = chaospy.poly.collection.core.variable(dimensions)
    if normed:
        orth = [
            poly**0*numpy.ones(dimensions),
            (poly-coeff1[:, 0])/numpy.sqrt(coeff2[:, 1]),
        ]
        for order_ in range(1, order):
            orth.append(
                (orth[-1]*(poly-coeff1[:, order_])
                 -orth[-2]*numpy.sqrt(coeff2[:, order_]))
                /numpy.sqrt(coeff2[:, order_+1])
            )
        norms = numpy.ones(coeff2.shape)
    else:
        orth = [poly-poly, poly**0*numpy.ones(dimensions)]
        for order_ in range(order):
            orth.append(
                orth[-1]*(poly-coeff1[:, order_])
                - orth[-2]*coeff2[:, order_]
            )
        orth = orth[1:]
        norms = numpy.cumprod(coeff2, 1)

    return orth, norms, coeff1, coeff2


def _stieltjes_approx(dist, order, accuracy, normed, **kws):
    """Stieltjes' method with approximative recurrence coefficients."""
    kws["rule"] = kws.get("rule", "C")
    assert kws["rule"].upper() != "G"

    absisas, weights = chaospy.quad.generate_quadrature(
        accuracy, dist.range(), **kws)
    weights = weights*dist.pdf(absisas)

    poly = chaospy.poly.variable(len(dist))
    orth = [poly*0, poly**0]

    inner = numpy.sum(absisas*weights, -1)
    norms = [numpy.ones(len(dist)), numpy.ones(len(dist))]
    coeff1 = []
    coeff2 = []

    for _ in range(order):

        coeff1.append(inner/norms[-1])
        coeff2.append(norms[-1]/norms[-2])
        orth.append((poly-coeff1[-1])*orth[-1] - orth[-2]*coeff2[-1])

        raw_nodes = orth[-1](*absisas)**2*weights
        inner = numpy.sum(absisas*raw_nodes, -1)
        norms.append(numpy.sum(raw_nodes, -1))

        if normed:
            orth[-1] = orth[-1]/numpy.sqrt(norms[-1])

    coeff1.append(inner/norms[-1])
    coeff2.append(norms[-1]/norms[-2])
    coeff1 = numpy.transpose(coeff1)
    coeff2 = numpy.transpose(coeff2)
    norms = numpy.array(norms[1:]).T
    orth = orth[1:]
    return orth, norms, coeff1, coeff2



"""
Uncertainty Quantification Toolbox
==================================

This module contains tools for performing uncertainty quantification of models.
"""
from chaospy.version import __version__
import chaospy.bertran
import chaospy.chol
import chaospy.descriptives
import chaospy.distributions
import chaospy.orthogonal
import chaospy.poly
import chaospy.spectral
import chaospy.quad
import chaospy.saltelli
import chaospy.regression

from chaospy.distributions import *
from chaospy.orthogonal import *
from chaospy.poly import *
from chaospy.spectral import *
from chaospy.quad import *
from chaospy.saltelli import *
from chaospy.descriptives import *
from chaospy.regression import *



"""
Negative of a distribution.

Example usage
-------------

Invert sign of a distribution::

    >>> distribution = -chaospy.Uniform(0, 1)
    >>> print(distribution)
    (-Uniform(0,1))
    >>> print(numpy.around(distribution.sample(5), 4))
    [-0.3464 -0.885  -0.0497 -0.5178 -0.1275]
    >>> print(distribution.fwd([-0.3, -0.2, -0.1]))
    [0.7 0.8 0.9]
    >>> print(distribution.inv(distribution.fwd([-0.3, -0.2, -0.1])))
    [-0.3 -0.2 -0.1]
    >>> print(distribution.pdf([-0.3, -0.2, -0.1]))
    [1. 1. 1.]
    >>> print(numpy.around(distribution.mom([1, 2, 3]), 4))
    [-0.5     0.3333 -0.25  ]
    >>> print(numpy.around(distribution.ttr([1, 2, 3]), 4))
    [[-0.5    -0.5    -0.5   ]
     [ 0.0833  0.0667  0.0643]]

"""
import numpy
from ..baseclass import Dist


class Neg(Dist):
    """Negative of a distribution."""

    def __init__(self, left):
        """
        Constructor.

        Args:
            left (Dist) : distribution.
        """
        Dist.__init__(self, left=left,
                _length=len(left), _advance=True)

    def _bnd(self, xloc, graph):
        """Distribution bounds."""
        return -graph(-xloc, graph.dists["left"])[::-1]

    def _pdf(self, xloc, graph):
        """Probability density function."""
        return graph(-xloc, graph.dists["left"])

    def _cdf(self, xloc, graph):
        """Cumulative distribution function."""
        return 1-graph(-xloc, graph.dists["left"])

    def _ppf(self, q, graph):
        """Point percentile function."""
        return -graph(1-q, graph.dists["left"])

    def _mom(self, k, graph):
        """Statistical moments."""
        return (-1)**numpy.sum(k, 0)*graph(k, graph.dists["left"])

    def _ttr(self, k, graph):
        """Three terms recursion coefficients."""
        a,b = graph(k, graph.dists["left"])
        return -a,b

    def _str(self, left):
        """String representation."""
        return "(-" + str(left) + ")"

    def _val(self, graph):
        """Value extraction."""
        if "left" in graph.keys:
            return -graph.keys["left"]
        return self

def neg(left):
    """
    Negative of a distribution.

    Args:
        dist (Dist) : distribution.
    """
    if not isinstance(left, Dist):
        return -left
    return Neg(left)




"""
Clenshaw-Curtis quadrature method.
"""
from __future__ import division

import numpy
import chaospy.quad


def quad_clenshaw_curtis(
        order, lower=0, upper=1, growth=False, composite=1, part=None):
    """
    Generate the quadrature nodes and weights in Clenshaw-Curtis quadrature.

    Example:
        >>> abscissas, weights = quad_clenshaw_curtis(3, 0, 1)
        >>> print(numpy.around(abscissas, 4))
        [[0.   0.25 0.75 1.  ]]
        >>> print(numpy.around(weights, 4))
        [0.1111 0.3889 0.1944 0.1111]
    """
    order = numpy.asarray(order, dtype=int).flatten()
    lower = numpy.asarray(lower).flatten()
    upper = numpy.asarray(upper).flatten()

    dim = max(lower.size, upper.size, order.size)

    order = numpy.ones(dim, dtype=int)*order
    lower = numpy.ones(dim)*lower
    upper = numpy.ones(dim)*upper

    if isinstance(composite, int):
        composite = numpy.array([numpy.linspace(0, 1, composite+1)]*dim)

    else:
        composite = numpy.asarray(composite)
        if not composite.shape:
            composite = composite.flatten()
        if len(composite.shape) == 1:
            composite = numpy.array([composite])
        composite = ((composite.T-lower)/(upper-lower)).T

    if growth:
        results = [
            _clenshaw_curtis(2**order[i]-1*(order[i] == 0), composite[i])
            for i in range(dim)
        ]
    else:
        results = [
            _clenshaw_curtis(order[i], composite[i]) for i in range(dim)
        ]

    abscis = [_[0] for _ in results]
    weight = [_[1] for _ in results]

    abscis = chaospy.quad.combine(abscis, part=part).T
    weight = chaospy.quad.combine(weight, part=part)

    abscis = ((upper-lower)*abscis.T + lower).T
    weight = numpy.prod(weight*(upper-lower), -1)

    assert len(abscis) == dim
    assert len(weight) == len(abscis.T)

    return abscis, weight


def _clenshaw_curtis(order, composite=None):
    """Backend method."""
    if order == 0:
        return numpy.array([.5]), numpy.array([1.])

    abscis = -numpy.cos(numpy.arange(order+1)*numpy.pi/order)
    abscis[numpy.abs(abscis) < 1e-14] = 0

    grid = numpy.meshgrid(*[numpy.arange(order//2+1)]*2)
    premat = 2./order*numpy.cos(2*grid[0]*grid[1]*numpy.pi/order)
    premat[:, 0] *= .5
    premat[:, -1] *= .5

    prevec = 2./(1-numpy.arange(0, order+1, 2)**2)
    prevec[0] *= .5
    prevec[-1] *= .5

    weight = numpy.dot(premat.T, prevec)
    weight = numpy.concatenate((weight, weight[-1-1*(order%2 == 0)::-1]))
    weight[order // 2] *= 2

    abscis = .5*abscis+.5
    weight *= .5

    length = len(abscis)

    if composite is None:
        composite = []
    composite = list(set(composite))
    composite = [c for c in composite if (c < 1) and (c > 0)]
    composite.sort()
    composite = [0] + composite + [1]

    abscissas = numpy.zeros((length-1)*(len(composite)-1)+1)
    weights = numpy.zeros((length-1)*(len(composite)-1)+1)
    for dim in range(len(composite)-1):
        abscissas[dim*length-dim:(dim+1)*length-dim] = \
                abscis*(composite[dim+1]-composite[dim]) + composite[dim]
        weights[dim*length-dim:(dim+1)*length-dim] += \
            weight*(composite[dim+1]-composite[dim])

    return abscissas, weights



"""
First order statistics functions.
"""
from itertools import product
import numpy

from .. import distributions, poly as polynomials, quad as quadrature


def E(poly, dist=None, **kws):
    """
    Expected value operator.

    1st order statistics of a probability distribution or polynomial on a given
    probability space.

    Args:
        poly (Poly, Dist) : Input to take expected value on.
        dist (Dist) : Defines the space the expected value is taken on.
                It is ignored if `poly` is a distribution.
        **kws (optional) : Extra keywords passed to dist.mom.

    Returns:
        (ndarray) : The expected value of the polynomial or distribution, where
                `expected.shape==poly.shape`.

    Examples:
        >>> x = chaospy.variable()
        >>> Z = chaospy.Uniform()
        >>> print(chaospy.E(Z))
        0.5
        >>> print(chaospy.E(x**3, Z))
        0.25
    """
    if not isinstance(poly, (distributions.Dist, polynomials.Poly)):
        print(type(poly))
        print("Approximating expected value...")
        out = quadrature.quad(poly, dist, veceval=True, **kws)
        print("done")
        return out

    if isinstance(poly, distributions.Dist):
        dist, poly = poly, polynomials.variable(len(poly))

    if not poly.keys:
        return numpy.zeros(poly.shape, dtype=int)

    if isinstance(poly, (list, tuple, numpy.ndarray)):
        return [E(_, dist, **kws) for _ in poly]

    if poly.dim < len(dist):
        poly = polynomials.setdim(poly, len(dist))

    shape = poly.shape
    poly = polynomials.flatten(poly)

    keys = poly.keys
    mom = dist.mom(numpy.array(keys).T, **kws)
    A = poly.A

    if len(dist) == 1:
        mom = mom[0]

    out = numpy.zeros(poly.shape)
    for i in range(len(keys)):
        out += A[keys[i]]*mom[i]

    out = numpy.reshape(out, shape)
    return out


def E_cond(poly, freeze, dist, **kws):

    assert not dist.dependent()

    if poly.dim < len(dist):
        poly = polynomials.setdim(poly, len(dist))

    freeze = polynomials.Poly(freeze)
    freeze = polynomials.setdim(freeze, len(dist))
    keys = freeze.keys
    if len(keys)==1 and keys[0]==(0,)*len(dist):
        freeze = list(freeze.A.values())[0]
    else:
        freeze = numpy.array(keys)
    freeze = freeze.reshape(int(freeze.size/len(dist)), len(dist))

    shape = poly.shape
    poly = polynomials.flatten(poly)

    kmax = numpy.max(poly.keys, 0) + 1
    keys = [range(k) for k in kmax]

    A = poly.A.copy()
    keys = poly.keys
    out = {}
    zeros = [0]*poly.dim

    for i in range(len(keys)):

        key = list(keys[i])
        a = A[tuple(key)]

        for d in range(poly.dim):
            for j in range(len(freeze)):
                if freeze[j, d]:
                    key[d], zeros[d] = zeros[d], key[d]
                    break

        tmp = a*dist.mom(tuple(key))
        if tuple(zeros) in out:
            out[tuple(zeros)] = out[tuple(zeros)] + tmp
        else:
            out[tuple(zeros)] = tmp

        for d in range(poly.dim):
            for j in range(len(freeze)):
                if freeze[j, d]:
                    key[d], zeros[d] = zeros[d], key[d]
                    break

    out = polynomials.Poly(out, poly.dim, poly.shape, float)
    out = polynomials.reshape(out, shape)

    return out



"""
Generate Chebyshev pseudo-random samples.

Example usage
-------------

Basic usage::

    >>> print(create_chebyshev_samples(order=2))
    [[0.25 0.75]]
    >>> print(numpy.around(create_chebyshev_samples(order=5), 4))
    [[0.067 0.25  0.5   0.75  0.933]]

Certain orders are nested::

    >>> print(numpy.around(create_chebyshev_samples(order=3), 4))
    [[0.1464 0.5    0.8536]]
    >>> print(numpy.around(create_chebyshev_samples(order=7), 4))
    [[0.0381 0.1464 0.3087 0.5    0.6913 0.8536 0.9619]]

Create nested samples directly with the dedicated function::

    >>> print(numpy.around(create_nested_chebyshev_samples(order=2), 4))
    [[0.1464 0.5    0.8536]]
    >>> print(numpy.around(create_nested_chebyshev_samples(order=3), 4))
    [[0.0381 0.1464 0.3087 0.5    0.6913 0.8536 0.9619]]

Multivariate usage::

    >>> print(numpy.around(create_chebyshev_samples(order=2, dim=2), 4))
    [[0.25 0.25 0.75 0.75]
     [0.25 0.75 0.25 0.75]]
"""
import numpy

import chaospy.quad


def create_chebyshev_samples(order, dim=1):
    """
    Chebyshev sampling function.

    Args:
        order (int):
            The number of samples to create along each axis.
        dim (int):
            The number of dimensions to create samples for.

    Returns:
        samples following Chebyshev sampling scheme mapped to the
        ``[0, 1]^dim`` hyper-cube and ``shape == (dim, order)``.
    """
    x_data = .5*numpy.cos(numpy.arange(order, 0, -1)*numpy.pi/(order+1)) + .5
    x_data = chaospy.quad.combine([x_data]*dim)
    return x_data.T


def create_nested_chebyshev_samples(order, dim=1):
    """
    Nested Chebyshev sampling function.

    Args:
        order (int):
            The number of samples to create along each axis.
        dim (int):
            The number of dimensions to create samples for.

    Returns:
        samples following Chebyshev sampling scheme mapped to the
        ``[0, 1]^dim`` hyper-cube and ``shape == (dim, 2**order-1)``.
    """
    return create_chebyshev_samples(order=2**order-1, dim=dim)



"""
Laja quadrature.

After paper by Narayan and Jakeman.
"""
import numpy
from scipy.optimize import fminbound

import chaospy.quad


def quad_leja(order, dist):
    """
    Generate Leja quadrature node.

    Example:
        >>> abscisas, weights = quad_leja(3, chaospy.Normal(0, 1))
        >>> print(numpy.around(abscisas, 4))
        [[-2.7173 -1.4142  0.      1.7635]]
        >>> print(numpy.around(weights, 4))
        [0.022  0.1629 0.6506 0.1645]
    """
    assert not dist.dependent()

    if len(dist) > 1:
        if isinstance(order, int):
            out = [quad_leja(order, _) for _ in dist]
        else:
            out = [quad_leja(order[_], dist[_]) for _ in range(len(dist))]

        abscissas = [_[0][0] for _ in out]
        weights = [_[1] for _ in out]
        abscissas = chaospy.quad.combine(abscissas).T
        weights = chaospy.quad.combine(weights)
        weights = numpy.prod(weights, -1)

        return abscissas, weights

    lower, upper = dist.range()
    abscissas = [lower, dist.mom(1), upper]
    for _ in range(int(order)):

        obj = create_objective(dist, abscissas)
        opts, vals = zip(
            *[fminbound(
                obj, abscissas[idx], abscissas[idx+1], full_output=1)[:2]
              for idx in range(len(abscissas)-1)]
        )
        index = numpy.argmin(vals)
        abscissas.insert(index+1, opts[index])

    abscissas = numpy.asfarray(abscissas).flatten()[1:-1]
    weights = create_weights(abscissas, dist)
    abscissas = abscissas.reshape(1, abscissas.size)

    return numpy.array(abscissas), numpy.array(weights)


def create_objective(dist, abscissas):
    """Create objective function."""
    abscissas_ = numpy.array(abscissas[1:-1])
    def obj(absisa):
        """Local objective function."""
        out = -numpy.sqrt(dist.pdf(absisa))
        out *= numpy.prod(numpy.abs(abscissas_ - absisa))
        return out
    return obj


def create_weights(nodes, dist):
    """Create weights for the Laja method."""
    poly = chaospy.quad.generate_stieltjes(dist, len(nodes)-1, retall=True)[0]
    poly = chaospy.poly.flatten(chaospy.poly.Poly(poly))
    weights_inverse = poly(nodes)
    weights = numpy.linalg.inv(weights_inverse)
    return weights[:, 0]



"""Second order statistics."""
import numpy
from scipy.stats import spearmanr

from .. import distributions, poly as polynomials


def Cov(poly, dist=None, **kws):
    """
    Covariance matrix, or 2rd order statistics.

    Args:
        poly (Poly, Dist) : Input to take covariance on. Must have
                `len(poly)>=2`.
        dist (Dist) : Defines the space the covariance is taken on.  It is
                ignored if `poly` is a distribution.
        **kws (optional) : Extra keywords passed to dist.mom.

    Returns:
        (ndarray) : Covariance matrix with
                `covariance.shape==poly.shape+poly.shape`.

    Examples:
        >>> Z = chaospy.MvNormal([0, 0], [[2, .5], [.5, 1]])
        >>> print(numpy.around(chaospy.Cov(Z), 4))
        [[2.  0.5]
         [0.5 1. ]]

        >>> x = chaospy.variable()
        >>> Z = chaospy.Normal()
        >>> print(numpy.around(chaospy.Cov([x, x**2], Z), 4))
        [[1. 0.]
         [0. 2.]]
    """
    if not isinstance(poly, (distributions.Dist, polynomials.Poly)):
        poly = polynomials.Poly(poly)

    if isinstance(poly, distributions.Dist):
        x = polynomials.variable(len(poly))
        poly, dist = x, poly
    else:
        poly = polynomials.Poly(poly)

    dim = len(dist)
    shape = poly.shape
    poly = polynomials.flatten(poly)
    keys = poly.keys
    N = len(keys)
    A = poly.A
    keys1 = numpy.array(keys).T
    if dim==1:
        keys1 = keys1[0]
        keys2 = sum(numpy.meshgrid(keys, keys))
    else:
        keys2 = numpy.empty((dim, N, N))
        for i in range(N):
            for j in range(N):
                keys2[:, i, j] = keys1[:, i]+keys1[:, j]

    m1 = dist.mom(keys1, **kws)
    m2 = dist.mom(keys2, **kws)
    mom = m2-numpy.outer(m1, m1)

    out = numpy.zeros((len(poly), len(poly)))
    for i in range(len(keys)):
        a = A[keys[i]]
        out += numpy.outer(a, a)*mom[i, i]
        for j in range(i+1, len(keys)):
            b = A[keys[j]]
            ab = numpy.outer(a, b)
            out += (ab+ab.T)*mom[i, j]

    out = numpy.reshape(out, shape+shape)
    return out



def Corr(poly, dist=None, **kws):
    """
    Correlation matrix of a distribution or polynomial.

    Args:
        poly (Poly, Dist) : Input to take correlation on. Must have
                `len(poly)>=2`.
        dist (Dist) : Defines the space the correlation is taken on.  It is
                ignored if `poly` is a distribution.
        **kws (optional) : Extra keywords passed to dist.mom.

    Returns:
        (ndarray) : Correlation matrix with
                `correlation.shape==poly.shape+poly.shape`.

    Examples:
        >>> Z = chaospy.MvNormal([3, 4], [[2, .5], [.5, 1]])
        >>> print(numpy.around(chaospy.Corr(Z), 4))
        [[1.     0.3536]
         [0.3536 1.    ]]

        >>> x = chaospy.variable()
        >>> Z = chaospy.Normal()
        >>> print(numpy.around(chaospy.Corr([x, x**2], Z), 4))
        [[1. 0.]
         [0. 1.]]
    """
    if isinstance(poly, distributions.Dist):
        poly, dist = polynomials.variable(len(poly)), poly
    else:
        poly = polynomials.Poly(poly)

    cov = Cov(poly, dist, **kws)
    var = numpy.diag(cov)
    vvar = numpy.sqrt(numpy.outer(var, var))
    return numpy.where(vvar > 0, cov/vvar, 0)



# pylint: disable=protected-access
"""Probability density function."""
import numpy

from ... import approx

def pdf_call(self, x_data, dist):
    "PDF call backend wrapper"
    self.counting(dist, "pdf")
    assert x_data.shape == (len(dist), self.size)
    self.dist, dist_ = dist, self.dist

    graph = self.graph
    graph.add_node(dist, key=x_data)
    out = numpy.empty(x_data.shape)

    prm = self.dists.build()
    prm.update(self.keys.build())
    for key, value in prm.items():
        if not isinstance(value, numpy.ndarray):
            value_ = self.run(value, "val")[0]
            if isinstance(value_, numpy.ndarray):
                prm[key] = value_
                graph.add_node(value, key=value_)

    if hasattr(dist, "_pdf"):
        if dist.advance:
            out[:] = dist._pdf(x_data, self)
        else:
            out[:] = dist._pdf(x_data, **prm)
    else:
        out = approx.pdf(dist, x_data, self, **self.meta)
    graph.add_node(dist, val=out)

    self.dist = dist_
    return numpy.array(out)



"""
Basic tools for Bertran index manipulation.
"""
import functools

import numpy
import scipy.misc

import chaospy.bertran


def add(idxi, idxj, dim):
    """
    Bertran addition.

    Example
    -------
    >>> print(chaospy.bertran.add(3, 3, 1))
    6
    >>> print(chaospy.bertran.add(3, 3, 2))
    10
    """
    idxm = numpy.array(multi_index(idxi, dim))
    idxn = numpy.array(multi_index(idxj, dim))
    out = single_index(idxm + idxn)
    return out


def terms(order, dim):
    """
    Count the number of polynomials in an expansion.

    Parameters
    ----------
    order : int
        The upper order for the expansion.
    dim : int
        The number of dimensions of the expansion.

    Returns
    -------
    N : int
        The number of terms in an expansion of upper order `M` and
        number of dimensions `dim`.
    """
    return int(scipy.misc.comb(order+dim, dim, 1))


def multi_index(idx, dim):
    """
    Single to multi-index using graded reverse lexicographical notation.

    Parameters
    ----------
    idx : int
        Index in interger notation
    dim : int
        The number of dimensions in the multi-index notation

    Returns
    -------
    out : tuple
        Multi-index of `idx` with `len(out)=dim`

    Examples
    --------
    >>> for idx in range(5):
    ...     print(chaospy.bertran.multi_index(idx, 3))
    (0, 0, 0)
    (1, 0, 0)
    (0, 1, 0)
    (0, 0, 1)
    (2, 0, 0)

    See Also
    --------
    single_index
    """
    def _rec(idx, dim):
        idxn = idxm = 0
        if not dim:
            return ()

        if idx == 0:
            return (0, )*dim

        while terms(idxn, dim) <= idx:
            idxn += 1
        idx -= terms(idxn-1, dim)

        if idx == 0:
            return (idxn,) + (0,)*(dim-1)
        while terms(idxm, dim-1) <= idx:
            idxm += 1

        return (int(idxn-idxm),) + _rec(idx, dim-1)

    return _rec(idx, dim)


def bindex(start, stop=None, dim=1, sort="G", cross_truncation=1.):
    """
    Generator for creating multi-indices.

    Parameters
    ----------
    start : int
        The lower order of the indices
    stop : int, optional
        the maximum shape included. If omitted:
        stop <- start; start <- 0
        If int is provided, set as largest total order.
        If array of int, set as largest order along each axis.
    dim : int
        The number of dimensions in the expansion
    cross_truncation : float
        Use hyperbolic cross truncation scheme to reduce the number of
        terms in expansion.

    Returns
    -------
    indices : list
        Grevlex order list of indices.

    Examples
    --------
    >>> print(chaospy.bertran.bindex(0, 1, 3))
    [(0, 0, 0), (1, 0, 0), (0, 1, 0), (0, 0, 1)]
    >>> print(chaospy.bertran.bindex(2, 3, 2))
    [(2, 0), (1, 1), (0, 2), (3, 0), (2, 1), (1, 2), (0, 3)]
    """
    if stop is None:
        start, stop = 0, start
    sort = sort.upper()

    local, total = [], []
    for idx in range(
            single_index([start] + [0]*(dim-1)),
            single_index([0]*(dim-1) + [start])+1):
        idxm = multi_index(idx, dim)
        local.append(list(idxm))
        total.append(idxm)

    for _ in range(start, stop):

        local_, local = local, []
        for idxm in local_:

            idxi = 0
            for idxi in range(len(idxm)):
                if sum(idxm[idxi+1:]) == 0:
                    break

            for idxj in range(idxi, len(idxm)):
                idxm[idxj] += 1
                local.append(idxm[:])
                total.append(tuple(idxm))
                idxm[idxj] -= 1

    if "G" in sort:
        total = sorted(total, key=sum)

    else:
        def cmp_(idxi, idxj):
            """Old style compare method."""
            if not numpy.any(idxi):
                return 0
            if idxi[0] == idxj[0]:
                return cmp(idxi[:-1], idxj[:-1])
            return (idxi[-1] > idxj[-1]) - (idxi[-1] < idxj[-1])
        key = functools.cmp_to_key(cmp_)
        total = sorted(total, key=key)

    if "I" in sort:
        total = total[::-1]

    if "R" in sort:
        total = [idx[::-1] for idx in total]

    for pos, idx in reversed(list(enumerate(total))):
        idx = numpy.array(idx)
        if numpy.any(numpy.sum(idx**(1./cross_truncation)) > stop**(1./cross_truncation)):
            del total[pos]


    return total


def single_index(idxm):
    """
    Multi-index to single integer notation.

    Uses graded reverse lexicographical notation.

    Parameters
    ----------
    idxm : array_like
        Index in multi-index notation

    Returns
    -------
    idx : int
        Integer index of `idxm`

    Examples
    --------
    >>> for idx in range(3):
    ...     print(chaospy.bertran.single_index(numpy.eye(3)[idx]))
    1
    2
    3
    """
    if -1 in idxm:
        return 0
    order = int(sum(idxm))
    dim = len(idxm)
    if order == 0:
        return 0
    return terms(order-1, dim) + single_index(idxm[1:])


def rank(idx, dim):
    """Calculate the index rank according to Bertran's notation."""
    idxm = multi_index(idx, dim)
    out = 0
    while idxm[-1:] == (0,):
        out += 1
        idxm = idxm[:-1]
    return out


def parent(idx, dim, axis=None):
    """
    Parent node according to Bertran's notation.

    Parameters
    ----------
    idx : int
        Index of the child node.
    dim : int
        Dimensionality of the problem.
    axis : int
        Assume axis direction.

    Returns
    -------
    out : int
        Index of parent node with `j<=i`, and `j==i` iff `i==0`.
    axis : int
        Dimension direction the parent was found.
    """
    idxm = multi_index(idx, dim)
    if axis is None:
        axis = dim - numpy.argmin(1*(numpy.array(idxm)[::-1] == 0))-1

    if not idx:
        return idx, axis

    if idxm[axis] == 0:
        idxi = parent(parent(idx, dim)[0], dim)[0]
        while child(idxi+1, dim, axis) < idx:
            idxi += 1
        return idxi, axis

    out = numpy.array(idxm) - 1*(numpy.eye(dim)[axis])
    return single_index(out), axis


def child(idx, dim, axis):
    """
    Child node according to Bertran's notation.

    Parameters
    ----------
    idx : int
        Index of the parent node.
    dim : int
        Dimensionality of the problem.
    axis : int
        Dimension direction to define a child.
        Must have `0<=axis<dim`

    Returns
    -------
    out : int
        Index of child node with `out > idx`.

    Examples
    --------
    >>> print(chaospy.bertran.child(4, 1, 0))
    5
    >>> print(chaospy.bertran.child(4, 2, 1))
    8
    """
    idxm = multi_index(idx, dim)
    out = numpy.array(idxm) + 1*(numpy.eye(len(idxm))[axis])
    return single_index(out)


def olindex(order, dim):
    """
    Create an lexiographical sorted basis for a given order.

    Examples
    --------
    >>> chaospy.bertran.olindex(3, 2)
    array([[0, 3],
           [1, 2],
           [2, 1],
           [3, 0]])
    """
    idxm = [0]*dim
    out = []

    def _olindex(idx):
        """Recursive backend for olindex."""
        if numpy.sum(idxm) == order:
            out.append(idxm[:])
            return

        if idx == dim:
            return

        idxm_sum = numpy.sum(idxm)
        idx_saved = idxm[idx]

        for idxi in range(order - numpy.sum(idxm) + 1):

            idxm[idx] = idxi

            if idxm_sum < order:
                _olindex(idx+1)

            else:
                break
        idxm[idx] = idx_saved

    _olindex(0)
    return numpy.array(out)


def olindices(order, dim):
    """
    Create an lexiographical sorted basis for a given order.

    Examples:
        >>> chaospy.bertran.olindices(2, 2)
        array([[0, 0],
               [0, 1],
               [1, 0],
               [0, 2],
               [1, 1],
               [2, 0]])
    """
    indices = [olindex(o, dim) for o in range(order+1)]
    indices = numpy.vstack(indices)
    return indices



import numpy

from .. import distributions, poly as polynomials
from .first import E



def Skew(poly, dist=None, **kws):
    """
    Skewness operator.

    Element by element 3rd order statistics of a distribution or polynomial.

    Args:
        poly (Poly, Dist) : Input to take skewness on.
        dist (Dist) : Defines the space the skewness is taken on.
                It is ignored if `poly` is a distribution.
        **kws (optional) : Extra keywords passed to dist.mom.

    Returns:
        (ndarray) : Element for element variance along `poly`, where
                `skewness.shape==poly.shape`.

    Examples:
        >>> x = chaospy.variable()
        >>> Z = chaospy.Gamma()
        >>> print(chaospy.Skew(Z))
        2.0
    """
    if isinstance(poly, distributions.Dist):
        x = polynomials.variable(len(poly))
        poly, dist = x, poly
    else:
        poly = polynomials.Poly(poly)

    if poly.dim < len(dist):
        polynomials.setdim(poly, len(dist))

    shape = poly.shape
    poly = polynomials.flatten(poly)

    m1 = E(poly, dist)
    m2 = E(poly**2, dist)
    m3 = E(poly**3, dist)
    out = (m3-3*m2*m1+2*m1**3)/(m2-m1**2)**1.5

    out = numpy.reshape(out, shape)
    return out



def Kurt(poly, dist=None, fisher=True, **kws):
    """
    Kurtosis operator.

    Element by element 4rd order statistics of a distribution or polynomial.

    Args:
        poly (Poly, Dist) : Input to take kurtosis on.
        dist (Dist) : Defines the space the skewness is taken on.
                It is ignored if `poly` is a distribution.
        fisher (bool) : If True, Fisher's definition is used (Normal -> 0.0).
                If False, Pearson's definition is used (normal -> 3.0)
        **kws (optional) : Extra keywords passed to dist.mom.

    Returns:
        (ndarray) : Element for element variance along `poly`, where
                `skewness.shape==poly.shape`.

    Examples:
        >>> x = chaospy.variable()
        >>> Z = chaospy.Uniform()
        >>> print(numpy.around(chaospy.Kurt(Z), 8))
        -1.2
        >>> Z = chaospy.Normal()
        >>> print(numpy.around(chaospy.Kurt(x, Z), 8))
        0.0
    """
    if isinstance(poly, distributions.Dist):
        x = polynomials.variable(len(poly))
        poly, dist = x, poly
    else:
        poly = polynomials.Poly(poly)

    if fisher:
        adjust = 3
    else:
        adjust = 0

    shape = poly.shape
    poly = polynomials.flatten(poly)

    m1 = E(poly, dist)
    m2 = E(poly**2, dist)
    m3 = E(poly**3, dist)
    m4 = E(poly**4, dist)

    out = (m4-4*m3*m1 + 6*m2*m1**2 - 3*m1**4) /\
            (m2**2-2*m2*m1**2+m1**4) - adjust

    out = numpy.reshape(out, shape)
    return out



"""
Frontend for the Hermite Genz-Keister quadrature rule.
"""

import numpy as np
import scipy.special

import chaospy.quad


def quad_genz_keister(order, dist, rule=24):
    """
    Genz-Keister quadrature rule.

    Eabsicassample:
        >>> abscissas, weights = quad_genz_keister(
        ...         order=1, dist=chaospy.Uniform(0, 1))
        >>> print(numpy.around(abscissas, 4))
        [[0.0416 0.5    0.9584]]
        >>> print(numpy.around(weights, 4))
        [0.1667 0.6667 0.1667]
    """
    assert isinstance(rule, int)

    if len(dist) > 1:

        if isinstance(order, int):
            values = [quad_genz_keister(order, d, rule) for d in dist]
        else:
            values = [quad_genz_keister(order[i], dist[i], rule)
                      for i in range(len(dist))]

        abscissas = [_[0][0] for _ in values]
        abscissas = chaospy.quad.combine(abscissas).T
        weights = [_[1] for _ in values]
        weights = np.prod(chaospy.quad.combine(weights), -1)

        return abscissas, weights

    foo = chaospy.quad.genz_keister.COLLECTION[rule]
    abscissas, weights = foo(order)
    abscissas = dist.inv(scipy.special.ndtr(abscissas))
    abscissas = abscissas.reshape(1, abscissas.size)

    return abscissas, weights


if __name__ == "__main__":
    import doctest
    doctest.testmod()



"""
Implementation of probabilistic collocation method.

Here subsamples of the Golub-weightselsch method is removed at random and weights
renormalized.
"""
import numpy
import chaospy.quad


def probabilistic_collocation(order, dist, subset=.1):
    """
    Probabilistic collocation method.

    Args:
        order (int, array_like) : Quadrature order along each axis.
        dist (Dist) : Distribution to generate samples from.
        subset (float) : Rate of which to removed samples.
    """
    abscissas, weights = chaospy.quad.collection.golub_welsch(order, dist)

    likelihood = dist.pdf(abscissas)

    alpha = numpy.random.random(len(weights))
    alpha = likelihood > alpha*subset*numpy.max(likelihood)

    abscissas = abscissas.T[alpha].T
    weights = weights[alpha]
    return abscissas, weights



"""
Identify first argument passed to class Poly constructor.
"""

import numpy as np

import chaospy.poly.base
import chaospy.poly.dimension
import chaospy.poly.typing


def identify_core(core):
    """Identify the polynomial argument."""
    for datatype, identifier in {
            int: _identify_scaler,
            np.int8: _identify_scaler,
            np.int16: _identify_scaler,
            np.int32: _identify_scaler,
            np.int64: _identify_scaler,
            float: _identify_scaler,
            np.float16: _identify_scaler,
            np.float32: _identify_scaler,
            np.float64: _identify_scaler,
            chaospy.poly.base.Poly: _identify_poly,
            dict: _identify_dict,
            np.ndarray: _identify_iterable,
            list: _identify_iterable,
            tuple: _identify_iterable,
    }.items():
        if isinstance(core, datatype):
            return identifier(core)

    raise TypeError(
        "Poly arg: 'core' is not a valid type " + repr(core))


def _identify_scaler(core):
    """Specification for a scaler value."""
    return {(0,): np.asarray(core)}, 1, (), type(core)


def _identify_poly(core):
    """Specification for a polynomial."""
    return core.A, core.dim, core.shape, core.dtype


def _identify_dict(core):
    """Specification for a dictionary."""
    if not core:
        return {}, 1, (), int

    core = core.copy()
    key = sorted(core.keys(), key=chaospy.poly.base.sort_key)[0]
    shape = np.array(core[key]).shape
    dtype = np.array(core[key]).dtype
    dim = len(key)
    return core, dim, shape, dtype


def _identify_iterable(core):
    """Specification for a list, tuple, np.ndarray."""
    if isinstance(core, np.ndarray) and not core.shape:
        return {(0,):core}, 1, (), core.dtype

    core = [chaospy.poly.base.Poly(a) for a in core]
    shape = (len(core),) + core[0].shape

    dtype = chaospy.poly.typing.dtyping(*[_.dtype for _ in core])

    dims = np.array([a.dim for a in core])
    dim = np.max(dims)
    if dim != np.min(dims):
        core = [chaospy.poly.dimension.setdim(a, dim) for a in core]

    out = {}
    for idx, core_ in enumerate(core):

        for key in core_.keys:

            if not key in out:
                out[key] = np.zeros(shape, dtype=dtype)
            out[key][idx] = core_.A[key]

    return out, dim, shape, dtype



"""Miscellenious descriptive statistics."""
import numpy as np
from scipy.stats import spearmanr

from .second2d import Corr
from .. import distributions, poly as polynomials


def Acf(poly, dist, N=None, **kws):
    """
    Auto-correlation function.

    Args:
        poly (Poly) : Polynomial of interest. Must have `len(poly)>N`
        dist (Dist) : Defines the space the correlation is taken on.
        N (int, optional) : The number of time steps appart included. If omited
                set to `len(poly)/2+1`.
        **kws (optional) : Extra keywords passed to dist.mom.

    Returns:
        (ndarray) : Auto-correlation of `poly` with `shape=(N, )`.
                Note that by definition `Q[0]=1`.

    Examples:
        >>> poly = chaospy.prange(10)[1:]
        >>> Z = chaospy.Uniform()
        >>> print(numpy.around(chaospy.Acf(poly, Z, 5), 4))
        [1.     0.9915 0.9722 0.9457 0.9127]
    """
    if N is None:
        N = len(poly)/2 + 1

    corr = Corr(poly, dist, **kws)
    out = np.empty(N)

    for n in range(N):
        out[n] = np.mean(corr.diagonal(n), 0)

    return out


def Spearman(poly, dist, sample=10000, retall=False, **kws):
    """
    Calculate Spearman's rank-order correlation coefficient.

    Args:
        poly (Poly) : Polynomial of interest.
        dist (Dist) : Defines the space where correlation is taken.
        sample (int) : Number of samples used in estimation.
        retall (bool) : If true, return p-value as well.
        **kws (optional) : Extra keywords passed to dist.sample.

    Returns:
        rho (float or ndarray) : Correlation output. Of type float if
                two-dimensional problem. Correleation matrix if larger.
        p-value (float or ndarray) : The two-sided p-value for a hypothesis
                test whose null hypothesis is that two sets of data are
                uncorrelated, has same dimension as rho.
    """
    samples = dist.sample(sample, **kws)
    poly = polynomials.flatten(poly)
    Y = poly(*samples)
    if retall:
        return spearmanr(Y.T)
    return spearmanr(Y.T)[0]


def Perc(poly, q, dist, sample=10000, **kws):
    """
    Percentile function.

    Note that this function is an empirical function that operates using Monte
    Carlo sampling.

    Args:
        poly (Poly) : Polynomial of interest.
        q (array_like) : positions where percentiles are taken. Must be
                a number or an array, where all values are on the interval
                `[0, 100]`.
        dist (Dist) : Defines the space where percentile is taken.
        sample (int) : Number of samples used in estimation.
        **kws (optional) : Extra keywords passed to dist.sample.

    Returns:
        (ndarray) : Percentiles of `poly` with `Q.shape=poly.shape+q.shape`.

    Examples:
        >>> chaospy.seed(1000)
        >>> x, y = chaospy.variable(2)
        >>> poly = chaospy.Poly([x, x*y])
        >>> Z = chaospy.J(chaospy.Uniform(3, 6), chaospy.Normal())
        >>> print(chaospy.Perc(poly, [0, 50, 100], Z))
        [[  3.         -45.        ]
         [  4.5080777   -0.05862173]
         [  6.          45.        ]]
    """
    shape = poly.shape
    poly = polynomials.flatten(poly)

    q = np.array(q)/100.
    dim = len(dist)

    # Interior
    Z = dist.sample(sample, **kws)
    if dim==1:
        Z = (Z, )
        q = np.array([q])
    poly1 = poly(*Z)

    # Min/max
    mi, ma = dist.range().reshape(2, dim)
    ext = np.mgrid[(slice(0, 2, 1), )*dim].reshape(dim, 2**dim).T
    ext = np.where(ext, mi, ma).T
    poly2 = poly(*ext)
    poly2 = np.array([_ for _ in poly2.T if not np.any(np.isnan(_))]).T

    # Finish
    if poly2.shape:
        poly1 = np.concatenate([poly1, poly2], -1)
    samples = poly1.shape[-1]
    poly1.sort()
    out = poly1.T[np.asarray(q*(samples-1), dtype=int)]
    out = out.reshape(q.shape + shape)
    return out


def QoI_Dist(poly, dist, sample=10000, **kws):
    """
    Constructs distributions for the quantity of interests.

    The function constructs a kernel density estimator (KDE) for each
    polynomial (poly) by sampling it.  With the KDEs, distributions (Dists) are
    constructed.  The Dists can be used for e.g. plotting probability density
    functions (PDF), or to make a second uncertainty quantification simulation
    with that newly generated Dists.

    Args:
        poly (Poly) : Polynomial of interest.
        dist (Dist) : Defines the space where the samples for the KDE is taken
                from the poly.
        sample (int) : Number of samples used in estimation to construct the
                KDE.
        **kws (optional) : Extra keywords passed to dist.sample.

    Returns:
        (ndarray) : The constructed quantity of interest (QoI) distributions,
                where `qoi_dists.shape==poly.shape`.

    Examples:
        >>> chaospy.seed(1000)
        >>> dist = chaospy.Normal(0, 1)
        >>> x = chaospy.variable(1)
        >>> poly = chaospy.Poly([x])
        >>> qoi_dist = chaospy.QoI_Dist(poly, dist)
        >>> values = qoi_dist[0].pdf([-0.75, 0., 0.75])
        >>> print(np.around(values, 8))
        [0.29143037 0.39931708 0.29536329]
    """
    shape = poly.shape
    poly = polynomials.flatten(poly)
    dim = len(dist)

    #sample from the input dist
    samples = dist.sample(sample, **kws)

    qoi_dists = []
    for i in range(0, len(poly)):
        #sample the polynomial solution
        if dim == 1:
            dataset = poly[i](samples)
        else:
            dataset = poly[i](*samples)

        lo = dataset.min()
        up = dataset.max()

        #creates qoi_dist
        qoi_dist = distributions.SampleDist(dataset, lo, up)
        qoi_dists.append(qoi_dist)

    #reshape the qoi_dists to match the shape of the input poly
    qoi_dists = np.array(qoi_dists, distributions.Dist)
    qoi_dists = qoi_dists.reshape(shape)

    if not shape:
        qoi_dists = qoi_dists.item()

    return qoi_dists



"""
Generates samples from the `Sobol sequence`_.

Sobol sequences (also called LP_T sequences or (t, s) sequences in base 2) are
an example of quasi-random low-discrepancy sequences. They were first
introduced by the Russian mathematician Ilya M. Sobol in 1967.

These sequences use a base of two to form successively finer uniform partitions
of the unit interval and then reorder the coordinates in each dimension.

Example usage
-------------

Standard usage::

    >>> set_state(1000)
    >>> print(numpy.around(create_sobol_samples(order=2, dim=2), 4))
    [[0.2197 0.7197 0.9697]
     [0.0967 0.5967 0.3467]]
    >>> print(numpy.around(create_sobol_samples(order=2, dim=2), 4))
    [[0.4697 0.3447 0.8447]
     [0.8467 0.4717 0.9717]]
    >>> print(numpy.around(create_sobol_samples(order=2, dim=3), 4))
    [[0.5947 0.0947 0.0635]
     [0.2217 0.7217 0.1904]
     [0.9229 0.4229 0.0166]]
    >>> print(numpy.around(create_sobol_samples(order=2, dim=6), 4))
    [[0.5635 0.8135 0.3135]
     [0.6904 0.4404 0.9404]
     [0.5166 0.7666 0.2666]
     [0.1768 0.9268 0.4268]
     [0.0537 0.3037 0.8037]
     [0.2529 0.5029 0.0029]]

Licence
-------

This code is distributed under the GNU LGPL license.

The routine adapts the ideas of Antonov and Saleev. Original FORTRAN77 version
by Bennett Fox. MATLAB version by John Burkardt. PYTHON version by Corrado
Chisari.

Papers::

    Antonov, Saleev,
    USSR Computational Mathematics and Mathematical Physics,
    Volume 19, 1980, pages 252 - 256.

    Paul Bratley, Bennett Fox,
    Algorithm 659:
    Implementing Sobol's Quasirandom Sequence Generator,
    ACM Transactions on Mathematical Software,
    Volume 14, Number 1, pages 88-100, 1988.

    Bennett Fox,
    Algorithm 647:
    Implementation and Relative Efficiency of Quasirandom
    Sequence Generators,
    ACM Transactions on Mathematical Software,
    Volume 12, Number 4, pages 362-376, 1986.

    Ilya Sobol,
    USSR Computational Mathematics and Mathematical Physics,
    Volume 16, pages 236-242, 1977.

    Ilya Sobol, Levitan,
    The Production of Points Uniformly Distributed in a Multidimensional
    Cube (in Russian),
    Preprint IPM Akad. Nauk SSSR,
    Number 40, Moscow 1976.

.. Sobel sequence: https://en.wikipedia.org/wiki/Sobol_sequence
"""
import math

import numpy


RANDOM_SEED = 1
DIM_MAX = 40
LOG_MAX = 30

SOURCE_SAMPLES = numpy.zeros((DIM_MAX, LOG_MAX), dtype=int)
SOURCE_SAMPLES[0:40, 0] = 1
SOURCE_SAMPLES[2:40, 1] = (
    1, 3, 1, 3, 1, 3, 3, 1, 3, 1, 3, 1, 3, 1, 1, 3, 1, 3, 1, 3, 1, 3, 3, 1, 3,
    1, 3, 1, 3, 1, 1, 3, 1, 3, 1, 3, 1, 3
)
SOURCE_SAMPLES[3:40, 2] = (
    7, 5, 1, 3, 3, 7, 5, 5, 7, 7, 1, 3, 3, 7, 5, 1, 1, 5, 3, 3, 1, 7, 5, 1, 3,
    3, 7, 5, 1, 1, 5, 7, 7, 5, 1, 3, 3
)
SOURCE_SAMPLES[5:40, 3] = (
    1, 7, 9, 13, 11, 1, 3, 7, 9, 5, 13, 13, 11, 3, 15, 5, 3, 15, 7, 9, 13, 9,
    1, 11, 7, 5, 15, 1, 15, 11, 5, 3, 1, 7, 9
)
SOURCE_SAMPLES[7:40, 4] = (
    9, 3, 27, 15, 29, 21, 23, 19, 11, 25, 7, 13, 17, 1, 25, 29, 3, 31, 11, 5,
    23, 27, 19, 21, 5, 1, 17, 13, 7, 15, 9, 31, 9
)
SOURCE_SAMPLES[13:40, 5] = (
    37, 33, 7, 5, 11, 39, 63, 27, 17, 15, 23, 29, 3, 21, 13, 31, 25, 9, 49, 33,
    19, 29, 11, 19, 27, 15, 25
)
SOURCE_SAMPLES[19:40, 6] = (
    13, 33, 115, 41, 79, 17, 29, 119, 75, 73, 105, 7, 59, 65, 21, 3, 113, 61,
    89, 45, 107
)
SOURCE_SAMPLES[37:40, 7] = (7, 23, 39)
POLY = (
    1, 3, 7, 11, 13, 19, 25, 37, 59, 47, 61, 55, 41, 67, 97, 91, 109, 103,
    115, 131, 193, 137, 145, 143, 241, 157, 185, 167, 229, 171, 213, 191,
    253, 203, 211, 239, 247, 285, 369, 299
)


def set_state(seed_value=None, step=None):
    """Set random seed."""
    global RANDOM_SEED  # pylint: disable=global-statement
    if seed_value is not None:
        RANDOM_SEED = seed_value
    if step is not None:
        RANDOM_SEED += step


def create_sobol_samples(order, dim, seed=None):
    """
    Args:
        order (int):
            Number of unique samples to generate
        dim (int):
            Number of spacial dimensions. Must satisfy ``0 < dim < 41``.
        seed (int, optional):
            Starting seed. Non-positive values are treated as 1. If omitted,
            consequtive samples are used.

    Returns:
        quasi (numpy.ndarray):
            Quasi-random vector with ``shape == (dim, order+1)``.
    """
    assert 0 < dim < DIM_MAX, "dim in [1, 40]"

    # global RANDOM_SEED  # pylint: disable=global-statement
    # if seed is None:
    #     seed = RANDOM_SEED
    # RANDOM_SEED += order+1
    set_state(seed_value=seed)
    seed = RANDOM_SEED
    set_state(step=order+1)

    # Initialize row 1 of V.
    samples = SOURCE_SAMPLES.copy()
    maxcol = int(math.log(2**LOG_MAX-1, 2))+1
    samples[0, 0:maxcol] = 1

    # Initialize the remaining rows of V.
    for idx in range(1, dim):

        # The bits of the integer POLY(I) gives the form of polynomial:
        degree = int(math.log(POLY[idx], 2))

        #Expand this bit pattern to separate components:
        includ = numpy.array([val == "1" for val in bin(POLY[idx])[-degree:]])

        #Calculate the remaining elements of row I as explained
        #in Bratley and Fox, section 2.
        for idy in range(degree+1, maxcol+1):
            newv = samples[idx, idy-degree-1].item()
            base = 1
            for idz in range(1, degree+1):
                base *= 2
                if includ[idz-1]:
                    newv = newv ^ base * samples[idx, idy-idz-1].item()
            samples[idx, idy-1] = newv

    samples = samples[:dim]

    # Multiply columns of V by appropriate power of 2.
    samples *= 2**(numpy.arange(maxcol, 0, -1, dtype=int))

    #RECIPD is 1/(common denominator of the elements in V).
    recipd = 0.5**(maxcol+1)
    lastq = numpy.zeros(dim, dtype=int)

    seed = int(seed) if seed > 1 else 1

    for seed_ in range(seed):
        lowbit = len(bin(seed_)[2:].split("0")[-1])
        lastq[:] = lastq ^ samples[:, lowbit]

    #Calculate the new components of QUASI.
    quasi = numpy.empty((dim, order+1))
    for idx in range(order+1):
        lowbit = len(bin(seed+idx)[2:].split("0")[-1])
        quasi[:, idx] = lastq * recipd
        lastq[:] = lastq ^ samples[:, lowbit]

    return quasi



"""
Front end for generating psuedo-random samples.

All sampling methods:

Key     Name
------  ----------------
``C``   Chebyshev nodes
``NC``  Nested Chebyshev
``K``   Korobov
``R``   (Pseudo-)Random
``RG``  Regular grid
``NG``  Nested grid
``L``   Latin hypercube
``S``   Sobol
``H``   Halton
``M``   Hammersley
------  ----------------

Example usage
-------------

Generating simple samples::

    >>> print(numpy.around(generate_samples(order=4), 4))
    [[0.6536 0.115  0.9503 0.4822]]

Custom domain::

    >>> print(numpy.around(generate_samples(order=4, domain=[-1, 1]), 4))
    [[ 0.7449 -0.5753 -0.9186 -0.2056]]
    >>> print(numpy.around(generate_samples(order=4, domain=chaospy.Normal(0, 1)), 4))
    [[-0.7286  1.0016 -0.8166  0.651 ]]

Use a custom sampling scheme::

    >>> print(numpy.around(generate_samples(order=4, rule="H"), 4))
    [[0.75  0.125 0.625 0.375]]

Multivariate case::

    >>> print(numpy.around(generate_samples(order=4, domain=[[-1, 0], [0, 1]]), 4))
    [[-0.6078 -0.8177 -0.2565 -0.9304]
     [ 0.8853  0.9526  0.9311  0.4154]]
    >>> print(numpy.around(generate_samples(
    ...     order=4, domain=chaospy.J(chaospy.Normal(), chaospy.Uniform())), 4))
    [[-1.896   2.0975 -0.4135  0.5437]
     [ 0.3619  0.0351  0.8551  0.6573]]

Antithetic variates::

    >>> print(numpy.around(generate_samples(order=8, rule="H", antithetic=True), 4))
    [[0.75  0.25  0.125 0.875 0.625 0.375 0.375 0.625]]

Multivariate antithetic variates::

    >>> print(numpy.around(generate_samples(order=8, domain=2, rule="M", antithetic=True), 4))
    [[0.75  0.25  0.75  0.25  0.125 0.875 0.125 0.875]
     [0.25  0.25  0.75  0.75  0.5   0.5   0.5   0.5  ]]
"""
import logging
import numpy
from . import sequences, latin_hypercube

SAMPLERS = {
    "C": sequences.create_chebyshev_samples,
    "NC": sequences.create_nested_chebyshev_samples,
    "K": sequences.create_korobov_samples,
    "RG": sequences.create_grid_samples,
    "NG": sequences.create_nested_grid_samples,
    "S": sequences.create_sobol_samples,
    "H": sequences.create_halton_samples,
    "M": sequences.create_hammersley_samples,
    "L": latin_hypercube.create_latin_hypercube_samples,
    "R": lambda order, dim: numpy.random.random((dim, order)),
}


def generate_samples(order, domain=1, rule="R", antithetic=None):
    """
    Sample generator.

    Args:
        order (int):
            Sample order. Determines the number of samples to create.
        domain (Dist, int, array_like):
            Defines the space where the samples are generated. If integer is
            provided, the space ``[0, 1]^domain`` will be used. If array-like
            object is provided, a hypercube it defines will be used. If
            distribution, the domain it spans will be used.
        rule (str):
            rule for generating samples. The various rules are listed in
            :mod:`chaospy.distributions.sampler.generator`.
        antithetic (array_like, optional):
            List of bool. Represents the axes to mirror using antithetic
            variable.
    """
    logger = logging.getLogger(__name__)
    logger.debug("generating random samples using rule %s", rule)

    rule = rule.upper()

    if isinstance(domain, int):
        dim = domain
        trans = lambda x_data: x_data

    elif isinstance(domain, (tuple, list, numpy.ndarray)):
        domain = numpy.asfarray(domain)
        if len(domain.shape) < 2:
            dim = 1
        else:
            dim = len(domain[0])
        trans = lambda x_data: ((domain[1]-domain[0])*x_data.T + domain[0]).T

    else:
        dist = domain
        dim = len(dist)
        trans = dist.inv

    if antithetic is not None:

        from .antithetic import create_antithetic_variates
        antithetic = numpy.array(antithetic, dtype=bool).flatten()
        if antithetic.size == 1 and dim > 1:
            antithetic = numpy.repeat(antithetic, dim)

        size = numpy.sum(1*numpy.array(antithetic))
        order_saved = order
        order = int(numpy.log(order - dim))
        order = order if order > 1 else 1
        while order**dim < order_saved:
            order += 1
        trans_ = trans
        trans = lambda x_data: trans_(
            create_antithetic_variates(x_data, antithetic)[:, :order_saved])

    assert rule in SAMPLERS, "rule not recognised"
    sampler = SAMPLERS[rule]
    x_data = trans(sampler(order=order, dim=dim))

    logger.debug("order: %d, dim: %d -> shape: %s", order, dim, x_data.shape)
    return x_data



"""
Class for organizing dependencies in runtime.

Iteration with this element is required to create advanced
distributions.

To use a let a distribution be advanced, include the following
parameters on initiation::

    _advance=True       Initiates the advanced distribution
    _length=len(dist)   Te length of a distribution. Defaults to 1

Subsetting the Dist-class using the _advance feature requires the
following backend methods.
The required ones are::

    _cdf(self, x, graph)        Cumulative distribution function
    _bnd(self, x, graph)        Upper and lower bounds at location x

The following can be provided::

    _pdf(self, x, graph)        Probability density function
    _ppf(self, q, graph)        CDF inverse
    _mom(self, k, graph)        Statistical moment generator
    _ttr(self, k, graph)        TTR coefficients generator
    _dep(self, graph)           Dependency callback.
    _val(self, graph)           Value callback function. Used to
                            itterative check if a distribution has
                            been evaluated before, indirectly.
    _str(self, **prm)       Preaty print of distribution

Access parameter for distributions, input locations (or
constants), and function evaluations respectively::

    graph.dists, graph.keys, graph.values

See :mod:`dist.graph.containers` for usage.

Iterate through the graph by evaluating the distribution it is a wrapper
for::

    graph(z, dist)

where z is locations, and dist is the distribution it wraps.

Iterate through distribution's dependencies::

    for dist in graph

or get all distributions as a list::

    graph.dependencies

print the chaospy state of graph::

    print(graph)

Make a copy of chaospy state::

    graph.copy()

Initiate an itteration process::

    graph.run(...)

Switch from PDF to forward seemlessly::

    graph.fwd_as_pdf(x, dist)

From dist.baseclass we have that there are two new methods. They are
defined as follows::

    _val(self, graph)   If enough information exists to say that the
                    distribution is evaluated given the state of
                    parameters, return that value.  Else return self.

    _dep(self, graph)   Return a list l with len(l)==len(self) with all
                    distributions it depends upon as a set in each
                    element.  A common element between two sets implies
                    dependencies.  Use graph(dist) (instead of graph(x, dist)) to
                    generate a distributions dependencies.
"""
import numpy
import networkx

from . import calling, main, containers


def construct_graph(dist):
    graph = networkx.DiGraph()
    graph.add_node(dist)
    dist_collection = [dist]
    while dist_collection:

        cur_dist = dist_collection.pop()

        values = (value for value in cur_dist.prm.values()
                    if not isinstance(value, numpy.ndarray))
        for value in values:
            if value not in graph.node:
                graph.add_node(value)
                dist_collection.append(value)
            graph.add_edge(cur_dist, value)

    assert networkx.is_directed_acyclic_graph(graph)
    return graph


class Graph:
    """Backend engine for the distributions."""

    def __init__(self, dist):
        """
        Args:
            dist (Dist) : The root of the dependency tree.
        """
        graph = construct_graph(dist)

        self.graph_source = graph
        self.graph = graph

        self.valmode = False
        self.size = None
        self.meta = {}
        self.root = self.dist = dist

        # to be deprecated:
        self.dists = containers.Dists(self)
        self.keys = containers.Keys(self)
        self.values = containers.Values(self)

        self._call = None

    @property
    def dists_(self):
        """All distributions found in the parameters."""
        return {key: value for key, value in self.dist.prm.items()
                if not isinstance(value, numpy.ndarray)}

    @property
    def keys_(self):
        """All values, either constants or distribution substitutes."""
        return {key: value for key, value in self.dist.prm.items()
                if isinstance(value, numpy.ndarray)
                or "key" in self.graph.graph.node[value]}

    @property
    def values_(self):
        """Contains all evaluations of distributions."""
        return {key: value for key, value in self.dist.prm.items()
                if isinstance(value, numpy.ndarray)
                or "val" in self.graph.graph.node[value]}

    @property
    def dist_collection(self):
        """Create collection of all distribution in dependency graph."""
        return networkx.topological_sort(self.graph)

    def __call__(self, *args, **kwargs):
        return self._call(self, *args, **kwargs)

    def __str__(self):
        graph = self.graph
        out = ""
        for dist in self.dist_collection:
            out += str(dist) + "\n"
            for key in ["key", "val"]:
                if key in graph.node[dist]:
                    out += key + "\n"
                    out += str(graph.node[dist][key]) + "\n"
            for key in ["inv", "fwd", "pdf", "rnd", "range"]:
                if key in graph.node[dist]:
                    out += key + ": "
                    out += str(graph.node[dist][key]) + "\n"
            out += "\n"

        out = out[:-1]
        return out

    def __iter__(self):
        return self.dist_collection.__iter__()

    def copy(self):
        """Shallow copy of graph. Distribution stays the same."""
        graph = Graph(self.root)
        for node in self.graph.nodes():
            graph.graph.add_node(node, **self.graph.node[node])
        graph.graph.add_edges_from(self.graph.edges())

        graph.size = self.size
        graph.meta = self.meta
        graph._call = self._call

        return graph

    def run(self, x, mode, **meta):
        """Run through network to perform an operator."""
        return main.call(self, x, mode, **meta)

    def counting(self, dist, mode):
        """counter function. Used for meta analysis."""
        if dist in self.graph.node and \
                mode in self.graph.node[dist]:
            self.graph.node[dist][mode] += 1
        else:
            self.graph.add_node(dist, **{mode:1})

    def fwd_as_pdf(self, x, dist):
        """
        During a PDF-call, a switch to CDF might be necessary. This functions
        initiates this switch.
        """
        graph_source = self.graph_source
        graph = networkx.DiGraph()
        for node in graph_source.nodes():
            graph.add_node(node)
        graph.add_edges_from(graph_source.edges())
        graph, self.graph = self.graph, graph

        self._call = calling.fwd_call
        out = self(x, dist)

        graph, self.graph = self.graph, graph
        self._call = calling.pdf_call

        return out

    @property
    def dependencies(self):
        """Set of node dependencies."""
        return set(self.graph.nodes())



# pylint: disable=wildcard-import
r"""
As noted throughout the documentation, known distributions are created easily
by calling their name of the function of interest. For example to create
a Gaussian random variable::

   >>> distribution = chaospy.Normal(0,1)

To construct simple multivariate random variables with stochastically
independent components, either all the same using
:func:`~chaospy.distributions.joint.Iid`::

   >>> distribution = chaospy.Iid(chaospy.Normal(0, 1), 2)

Or with more detailed control through :func:`~chaospy.distributions.joint.J`::

   >>> distribution = chaospy.J(chaospy.Normal(0, 1), chaospy.Normal(0, 1))

The functionality of the distributions are covered in various other sections:

* To generate random samples, see :ref:`montecarlo`.
* To create transformations, see :ref:`rosenblatt`.
* To generate raw statistical moments, see :ref:`moments`.
* To generate three terms recurrence coefficients, see :ref:`orthogonality`.
* To analyse statistical properies, see :ref:`descriptives`.
"""
from . import baseclass
from .baseclass import Dist

from .constructor import construct

from . import (
    graph, sampler, approx, joint, cores,
    copulas, collection, operators, rosenblatt,
)

from .graph import *
from .sampler import *
from .approx import *
from .joint import *
from .cores import *
from .copulas import *
from .collection import *
from .operators import *

from numpy.random import seed



"""
Constructor of the Poly class.
"""
import chaospy.poly.constructor.identifier
import chaospy.poly.constructor.preprocessing
import chaospy.poly.constructor.ensurance

from chaospy.poly.constructor.identifier import *
from chaospy.poly.constructor.preprocessing import *
from chaospy.poly.constructor.ensurance import *



"""
Function for changing polynomial array's shape.
"""
import numpy as np

from chaospy.poly.base import Poly


def flatten(vari):
    """
    Flatten a shapeable quantity.

    Args:
        vari (Poly, array_like) : Shapeable input quantity.

    Returns:
        (Poly, array_like) : Same type as `vari` with `len(Q.shape)==1`.

    Examples:
        >>> P = cp.reshape(cp.prange(4), (2,2))
        >>> print(P)
        [[1, q0], [q0^2, q0^3]]
        >>> print(cp.flatten(P))
        [1, q0, q0^2, q0^3]
    """
    if isinstance(vari, Poly):
        shape = int(np.prod(vari.shape))
        return reshape(vari, (shape,))

    return np.array(vari).flatten()


def reshape(vari, shape):
    """
    Reshape the shape of a shapeable quantity.

    Args:
        vari (Poly, array_like) : Shapeable input quantity
        shape (tuple) : The polynomials new shape. Must be compatible with the
                number of elements in `vari`.

    Returns:
        (Poly, array_like) : Same type as `vari`.

    Examples:
        >>> poly = cp.prange(6)
        >>> print(poly)
        [1, q0, q0^2, q0^3, q0^4, q0^5]
        >>> print(cp.reshape(poly, (2,3)))
        [[1, q0, q0^2], [q0^3, q0^4, q0^5]]
    """

    if isinstance(vari, Poly):
        core = vari.A.copy()
        for key in vari.keys:
            core[key] = reshape(core[key], shape)
        out = Poly(core, vari.dim, shape, vari.dtype)
        return out

    return np.asarray(vari).reshape(shape)


def rollaxis(vari, axis, start=0):
    """
    Roll the specified axis backwards, until it lies in a given position.

    Args:
        vari (Poly, array_like) : Input array or polynomial.
        axis (int) : The axis to roll backwards. The positions of the
            other axes do not change realtie to one another.
        start (int, optional) : The axis is rolled until it lies before
            thes position.
    """
    if isinstance(vari, Poly):
        core_old = vari.A.copy()
        core_new = {}
        for key in vari.keys:
            core_new[key] = rollaxis(core_old[key], axis, start)
        return Poly(core_new, vari.dim, None, vari.dtype)

    return np.rollaxis(vari, axis, start)


def swapaxes(vari, ax1, ax2):
    """Interchange two axes of a polynomial."""
    if isinstance(vari, Poly):
        core = vari.A.copy()
        for key in vari.keys:
            core[key] = swapaxes(core[key], ax1, ax2)

        return Poly(core, vari.dim, None, vari.dtype)

    return np.swapaxes(vari, ax1, ax2)


def roll(vari, shift, axis=None):
    """Roll array elements along a given axis."""
    if isinstance(vari, Poly):
        core = vari.A.copy()
        for key in vari.keys:
            core[key] = roll(core[key], shift, axis)
        return Poly(core, vari.dim, None, vari.dtype)

    return np.roll(vari, shift, axis)


def transpose(vari):
    """
    Transpose a shapeable quantety.

    Args:
        vari (Poly, array_like) : Quantety of interest.

    Returns:
        Q (Poly, array_like) : Same type as `vari`.

    Examples:
        >>> P = cp.reshape(cp.prange(4), (2,2))
        >>> print(P)
        [[1, q0], [q0^2, q0^3]]

        >>> print(cp.transpose(P))
        [[1, q0^2], [q0, q0^3]]
    """
    if isinstance(vari, Poly):
        core = vari.A.copy()
        for key in vari.keys:
            core[key] = transpose(core[key])
        return Poly(core, vari.dim, vari.shape[::-1], vari.dtype)

    return np.transpose(vari)


import chaospy as cp

if __name__ == '__main__':
    import doctest
    doctest.testmod()



"""
Collection of hyerbolic operators

Example usage
-------------
"""
import numpy
from ..baseclass import Dist

class Sinh(Dist):

    def __init__(self, dist):
        Dist.__init__(self, dist=dist,
                _length=len(dist), _advance=True)

    def _str(self, dist):
        return "Sinh(%s)" % dist

    def _val(self, graph):
        if "dist" in graph.keys:
            return numpy.sinh(graph.keys["dist"])
        return self

    def _pdf(self, x, graph):
        return graph(numpy.arcsinh(x), graph.dists["dist"])/numpy.sqrt(1+x*x)

    def _cdf(self, x, graph):
        return graph(numpy.arcsinh(x), graph.dists["dist"])

    def _ppf(self, q, graph):
        return numpy.sinh(graph(q, graph.dists["dist"]))

    def _bnd(self, x, graph):
        return numpy.sinh(graph(numpy.arcsinh(x), graph.dists["dist"]))


class Arcsinh(Dist):

    def __init__(self, dist):
        Dist.__init__(self, dist=dist,
                _length=len(dist), _advance=True)

    def _str(self, dist):
        return "Arcsinh(%s)" % dist

    def _val(self, graph):
        if "dist" in graph.keys:
            return numpy.arcsinh(graph.keys["dist"])
        return self

    def _pdf(self, x, graph):
        return graph(numpy.sinh(x), graph.dists["dist"])*numpy.cosh(x)

    def _cdf(self, x, graph):
        return graph(numpy.sinh(x), graph.dists["dist"])

    def _ppf(self, q, graph):
        return numpy.arcsinh(graph(q, graph.dists["dist"]))

    def _bnd(self, x, graph):
        return numpy.arcsinh(graph(numpy.sinh(x), graph.dists["dist"]))


class Cosh(Dist):

    def __init__(self, dist):
        assert numpy.all(dist.range()>=0)
        Dist.__init__(self, dist=dist,
                _length=len(dist), _advance=True)

    def _str(self, dist):
        return "Cos(%s)" % dist

    def _val(self, graph):
        if "dist" in graph.keys:
            return numpy.cosh(graph.keys["dist"])
        return self

    def _pdf(self, x, graph):
        return graph(numpy.arccosh(x), graph.dists["dist"])/numpy.sqrt(x*x-1)

    def _cdf(self, x, graph):
        return graph(numpy.arccosh(x), graph.dists["dist"])

    def _ppf(self, q, graph):
        return numpy.cosh(graph(q, graph.dists["dist"]))

    def _bnd(self, x, graph):
        return numpy.cosh(graph(numpy.arccosh(x), graph.dists["dist"]))


class Arccosh(Dist):

    def __init__(self, dist):
        assert numpy.all(dist.range()>=1)
        Dist.__init__(self, dist=dist,
                _length=len(dist), _advance=True)

    def _str(self, dist):
        return "Arccosh(%s)" % dist

    def _val(self, graph):
        if "dist" in graph.keys:
            return numpy.arccosh(graph.keys["dist"])
        return self

    def _pdf(self, x, graph):
        return graph(numpy.cosh(x), graph.dists["dist"])*numpy.sinh(x)

    def _cdf(self, x, graph):
        return graph(numpy.cosh(x), graph.dists["dist"])

    def _ppf(self, q, graph):
        return numpy.arccosh(graph(q, graph.dists["dist"]))

    def _bnd(self, x, graph):
        return numpy.arccosh(graph(numpy.cosh(x), graph.dists["dist"]))


class Tanh(Dist):

    def __init__(self, dist):
        lo,up = dist.range()
        assert numpy.all(lo>=-1) and numpy.all(up<=1)
        Dist.__init__(self, dist=dist,
                _length=len(dist), _advance=True)

    def _str(self, dist):
        return "Tanh(%s)" % dist

    def _val(self, graph):
        if "dist" in graph.keys:
            return numpy.tanh(graph.keys["dist"])
        return self

    def _pdf(self, x, graph):
        return graph(numpy.arctanh(x), graph.dists["dist"])/numpy.sqrt(1-x*x)

    def _cdf(self, x, graph):
        return graph(numpy.arctanh(x), graph.dists["dist"])

    def _ppf(self, q, graph):
        return numpy.tanh(graph(q, graph.dists["dist"]))

    def _bnd(self, x, graph):
        return numpy.tanh(graph(numpy.arctanh(x), graph.dists["dist"]))


class Arctanh(Dist):

    def __init__(self, dist):
        Dist.__init__(self, dist=dist,
                _length=len(dist), _advance=True)

    def _str(self, dist):
        return "Arctanh(%s)" % dist

    def _val(self, graph):
        if "dist" in graph.keys:
            return numpy.arctanh(graph.keys["dist"])
        return self

    def _pdf(self, x, graph):
        return graph(numpy.tanh(x), graph.dists["dist"])/numpy.sinh(x)**2

    def _cdf(self, x, graph):
        return graph(numpy.tanh(x), graph.dists["dist"])

    def _ppf(self, q, graph):
        return numpy.arctanh(graph(q, graph.dists["dist"]))

    def _bnd(self, x, graph):
        return numpy.arctanh(graph(numpy.tanh(x), graph.dists["dist"]))



# pylint: disable=protected-access
"""Backend for raw statistical moment generator."""
import numpy

from ... import approx


def mom_call(self, expon, dist):
    "Moment generator call backend wrapper"
    assert len(expon) == len(dist)
    graph = self.graph
    self.dist, dist_ = dist, self.dist

    graph.add_node(dist, key=expon)
    if hasattr(dist, "_mom"):
        if dist.advance:
            out = dist._mom(expon, self)
        else:
            out = numpy.empty(expon.shape[1:])
            prm = self.dists.build()
            prm.update(self.keys.build())
            out[:] = dist._mom(expon, **prm)
    else:
        out = approx.mom(dist, expon, **self.meta)
    graph.add_node(dist, val=out)

    self.dist = dist_
    return numpy.array(out)



"""
Frontend collection.

``Gaussian``, ``G``
    Optimal Gaussian quadrature from Golub-Welsch. A slow method for higher order.
``Legendre``, ``E``
    Gauss-Legendre quadrature
``Clenshaw``, ``C``
    Clenshaw-Curtis quadrature. Exponential growth rule is used when sparse is
    True to make the rule nested.
``Leja``, J``
    Leja quadrature. Linear growth rule is nested.
``Genz``, ``Z``
    Hermite Genz-Keizter 16 rule. Nested. Valid to order 8.
``Patterson``, ``P``
    Gauss-Patterson quadrature rule. Nested. Valid to order 8.
"""
import inspect
import numpy

from .clenshaw_curtis import quad_clenshaw_curtis
from .gauss_patterson import quad_gauss_patterson
from .gauss_legendre import quad_gauss_legendre
from .genz_keister import quad_genz_keister
from .golub_welsch import quad_golub_welsch
from .leja import quad_leja

QUAD_FUNCTIONS = {
    "clenshaw_curtis": quad_clenshaw_curtis,
    "gauss_legendre": quad_gauss_legendre,
    "gauss_patterson": quad_gauss_patterson,
    "genz_keister": quad_genz_keister,
    "golub_welsch": quad_golub_welsch,
    "leja": quad_leja,
}

QUAD_SHORT_NAMES = {
    "c": "clenshaw_curtis",
    "e": "gauss_legendre",
    "p": "gauss_patterson",
    "z": "genz_keister",
    "g": "golub_welsch",
    "j": "leja",
}



def get_function(rule, domain, **parameters):
    """
    Create a quadrature function and set default parameter values.

    Args:
        rule (str) : Name of quadrature rule defined in `QUAD_FUNCTIONS`.
        domain (Dist, array_like) : Defines `lower` and `upper` that is passed
            quadrature rule. If `Dist`, `domain` is renamed to `dist` and also
            passed.
        **parameters (optional) : Redefining of the parameter defaults. Only
            add parameters that the quadrature rule expect.
    Returns:
        (callable) : Function that can be called only using argument `order`.
    """
    from ...distributions.baseclass import Dist
    if isinstance(domain, Dist):
        lower, upper = domain.range()
        parameters["dist"] = domain
    else:
        lower, upper = numpy.array(domain)
    parameters["lower"] = lower
    parameters["upper"] = upper

    quad_function = QUAD_FUNCTIONS[rule]
    parameters_spec = inspect.getargspec(quad_function)[0]
    parameters_spec = {key: None for key in parameters_spec}
    del parameters_spec["order"]

    for key in parameters_spec:
        if key in parameters:
            parameters_spec[key] = parameters[key]

    def _quad_function(order, *args, **kws):
        """Implementation of quadrature function."""
        params = parameters_spec.copy()
        params.update(kws)
        return quad_function(order, *args, **params)

    return _quad_function



"""
Logarithm base 10.
"""
import numpy

from ..baseclass import Dist


class Log10(Dist):
    """Logarithm base 10."""

    def __init__(self, dist):
        """
        Constructor.

        Args:
            dist (Dist) : distribution (>=0).
        """
        assert isinstance(dist, Dist)
        assert numpy.all(dist.range()>=0)
        Dist.__init__(self, dist=dist, _length=len(dist),
                _advance=True)

    def _str(self, dist):
        """String representation."""
        return "Log10(%s)" % dist

    def _val(self, graph):
        """Value extraction."""
        if "dist" in graph.keys:
            return numpy.log10(graph.keys["dist"])
        return self

    def _pdf(self, xloc, graph):
        """Probability density function."""
        return graph(10**xloc, graph.dists["dist"])*numpy.log(10)*10**xloc

    def _cdf(self, xloc, graph):
        """Cumulative distribution function."""
        return graph(10**xloc, graph.dists["dist"])

    def _ppf(self, q, graph):
        """Point percentile function."""
        return numpy.log10(graph(q, graph.dists["dist"]))

    def _bnd(self, xloc, graph):
        """Distribution bounds."""
        lower,upper = graph(10**xloc, graph.dists["dist"])
        return numpy.log10(lower), numpy.log10(upper)


def log10(dist):
    """
    Logarithm base 10.

    Args:
        dist (Dist) : distribution (>=0).
    """
    return Log10(dist)




"""
Convenience properties found on distribution object during dependency traversal.

Instead of iterating through each parameter in a distribution ``dist.prm``,
checking each what each state and acting from there, it is possible to access
a subset of parameters that fit certain criteria.

For example, to access all parameters that are also distributions, the property
``dist.graph.dists``::

    >>> dists0 = chaospy.beta(a=1, b=2).graph.dists
    >>> dists1 = chaospy.beta(a=chaospy.uniform(), b=2).graph.dists
    >>> dists2 = chaospy.beta(a=chaospy.uniform(), b=chaospy.gamma(1)).graph.dists

These objects can be used to identify how many distributions are present::

    >>> print(len(dists0), len(dists1), len(dists2))
    0 1 2

To identify specific parameters that are distributions::

    >>> print("a" in dists0, "a" in dists1, "a" in dists2)
    False True True
    >>> print("b" in dists0, "b" in dists1, "b" in dists2)
    False False True

An lastly to extract said parameters::

    >>> print(dists2["b"])
    gam(1)
    >>> print(dists1["a"])
    uni

Giving values not present or illegal values, causes exceptions::

    >>> dists0["b"]
    Traceback (most recent call last):
        ...
    KeyError: "parameter 'b' is not a distribution"
    >>> dists0["c"]
    Traceback (most recent call last):
        ...
    KeyError: "parameter key 'c' unknown; ('a', 'b') available"
"""
import numpy


class Container(object):
    """Baseclass for quick access attributes."""

    def __init__(self, graph):
        self.graph = graph

    def __contains__(self, key):
        raise NotImplementedError()

    def getitem(self, key):
        raise NotImplementedError()

    def __getitem__(self, key):
        if key not in self.graph.dist.prm:
            raise KeyError(
                "parameter key '%s' unknown; %s available" % (
                    key, tuple(self.graph.dist.prm)))
        return self.getitem(key)

    def build(self):
        "build a dict with all parameters"
        return {key: self[key] for key in self.graph.dist.prm if key in self}

    def __iter__(self):
        return self.build().values().__iter__()

    def __str__(self):
        return str(self.build())

    def __len__(self):
        return len(self.build())


class Dists(Container):
    """Contains all distributions."""

    def __contains__(self, key):
        return not isinstance(self.graph.dist.prm[key], numpy.ndarray)

    def getitem(self, key):
        if not key in self:
            raise KeyError("parameter '%s' is not a distribution" % key)
        return self.graph.dist.prm[key]


class Keys(Container):
    """
    Contains all values.

    Either as constants or as substitutes of distributions.
    """

    def __contains__(self, key):

        out = False
        val = self.graph.dist.prm[key]
        if isinstance(val, numpy.ndarray) or \
                "key" in self.graph.graph.node[val]:
            out = True

        return out

    def getitem(self, key):

        if not key in self:
            raise KeyError()
        val = self.graph.dist.prm[key]

        if isinstance(val, numpy.ndarray):
            return val

        gkey = self.graph.dist.prm[key]
        return self.graph.graph.node[gkey]["key"]

class Values(Container):
    """Contains all evaluations of distributions."""

    def __contains__(self, key):
        out = self.graph.dist.prm[key]
        return  not isinstance(out, numpy.ndarray) and \
                "val" in self.graph.graph.node[out]

    def getitem(self, key):
        if not key in self:
            raise KeyError()
        return self.graph.graph.node[self.graph.dist.prm[key]]["val"]



"""
Cunction to combine two dataset together with a tensor product.
"""

import numpy
import chaospy

def combine(args, part=None):
    """
    All linear combination of a list of list.

    Args:
        args (array_like) : List of input arrays.  Components to take linear
            combination of with `args[i].shape=(N[i], M[i])` where N is to be
            taken linear combination of and M is static.  M[i] is set to 1 if
            missing.

    Returns:
        (numpy.array) : matrix of combinations with shape (numpy.prod(N),
            numpy.sum(M)).

    Examples:
        >>> A, B = [1,2], [[4,4],[5,6]]
        >>> print(chaospy.quad.combine([A, B]))
        [[1. 4. 4.]
         [1. 5. 6.]
         [2. 4. 4.]
         [2. 5. 6.]]
    """
    args = [cleanup(arg) for arg in args]

    if part is not None:
        parts, orders = part
        if numpy.array(orders).size == 1:
            orders = [int(numpy.array(orders).item())]*len(args)
        parts = numpy.array(parts).flatten()

        for i, arg in enumerate(args):
            m, n = float(parts[i]), float(orders[i])
            l = len(arg)
            args[i] = arg[int(m/n*l):int((m+1)/n*l)]

    shapes = [arg.shape for arg in args]
    size = numpy.prod(shapes, 0)[0]*numpy.sum(shapes, 0)[1]

    if size > 10**9:
        raise MemoryError("Too large sets")

    if len(args) == 1:
        out = args[0]

    elif len(args) == 2:
        out = combine_two(*args)

    else:
        arg1 = combine_two(*args[:2])
        out = combine([arg1,]+args[2:])

    return out


def combine_two(arg1, arg2):
    l1, d1 = arg1.shape
    l2, d2 = arg2.shape
    out = numpy.empty((l1*l2, d1+d2))
    out[:,:d1] = numpy.tile(arg1, l2).reshape(l1*l2, d1)
    out[:,d1:] = numpy.tile(arg2.T, l1).reshape(d2, l1*l2).T
    return out


def cleanup(arg):
    """Clean up the input variable."""
    arg = numpy.asarray(arg)
    if len(arg.shape) <= 1:
        arg = arg.reshape(arg.size, 1)
    elif len(arg.shape) > 2:
        raise ValueError("shapes must be smaller than 3")
    return arg

if __name__=="__main__":
    import doctest
    doctest.testmod()



# pylint: disable=protected-access
"""Main operator."""
import networkx

from . import calling


def call(self, x_data, mode, **meta):
    """
    Run through network to perform an operator.

    Using this during another call is possible, but care has to be
    taken to ensure that the chaospy graph state is compatible with the
    new call.

    Available modes to run in:
    ------- -------------------------------------------
    "fwd"   Forward
    "inv"   Inverse
    "range" Lower and upper limits
    "ttr"   Three terms recursion coefficient generator
    "mom"   Raw statistical moments
    "dep"   Dependencies
    "val"   Values
    ------- -------------------------------------------

    Args:
        x_data (numpy.ndarray) : Input variable with shape=(D,K), where D is the
                number of dimensions in dist.
        mode (str) : What type of operator to perform.
    **meta
        Keyword argument passed to approximation method if run.
    """
    if mode in ("rnd", "dep", "val"):
        self.size, size = x_data, self.size
    else:
        self.size, size = x_data.shape[-1], self.size
    self.meta, meta = meta, self.meta

    if mode != "val":
        graph_source = self.graph_source
        graph = networkx.DiGraph()
        for node in graph_source.nodes():
            graph.add_node(node)
        graph.add_edges_from(graph_source.edges())
        graph, self.graph = self.graph, graph

    stored_call = self._call
    assert mode in calling.CALL_FUNCTIONS, "unknown mode %s" % mode
    self._call = calling.CALL_FUNCTIONS[mode]

    if mode != "val":
        self.dist = self.root

    if mode in ("rnd", "dep"):
        out = self(self.root)

    elif mode == "val":
        out = self(x_data)
    else:
        out = self(x_data, self.root)

    # if mode == "ttr":
    #     out[1] = out[1]**(x_data != 0)

    self.size = size
    if mode == "val":
        graph = self.graph
    else:
        graph, self.graph = self.graph, graph
    self._call = stored_call
    self.meta = meta

    return out, graph



"""
One of ``chaospy``'s most powerful features is possible to construct advance
multivariate variables directly through dependencies. To illustrate this,
consider the following bivariate distribution::

    >>> dist_ind = chaospy.Gamma(1)
    >>> dist_dep = chaospy.Normal(dist_ind**2, dist_ind+1)
    >>> distribution = chaospy.J(dist_ind, dist_dep)

In other words, a distribution dependent upon another distribution was created
simply by inserting it into the constructor of another distribution. The
resulting bivariate distribution if fully valid with dependent components.
For example the probability density function functions will in this case look
as follows::

    >>> x, y = np.meshgrid(np.linspace(-4, 7), np.linspace(0, 3))
    >>> likelihood = distribution.pdf([x, y])

This method also allows for construct any multivariate probability distribution
as long as you can fully construct the distribution's conditional decomposition
as noted in :ref:`rosenblatt`. One only has to construct each univariate
probability distribution and add dependencies in through the parameter
structure.

Now it is worth noting a couple of cavats:

* Since the underlying feature to accomplish this is the :ref:`rosenblatt`, the
  number of unique random variables in the final joint distribution has to be
  constant. In other words, ``dist_dep`` is not a valid distribution in itself,
  since it is univariat, but depends on the results of ``dist_ind``.
* The dependency order does not matter as long as it can defined as an acyclic
  graph. In other words, ``dist_ind`` can not be dependent upon ``dist_dep`` at
  the same time as ``dist_dep`` is dependent upon ``dist_ind``.
"""
from copy import copy

import numpy as np

from .baseclass import Dist


class Joint(Dist):
    """Joint probability operator."""

    def __init__(self, *args):
        """
Parameters
----------
*args : [Dist, ..., Dist]
    Set of univariate distributions to join into joint.
        """

        assert np.all([isinstance(_, Dist) for _ in args])
        prm = {"_%03d" % i:args[i] for i in range(len(args))}
        Dist.__init__(self, _advance=True, _length=len(args), **prm)
        self.sorting = []
        for dist in self.graph:
            if dist in args:
                self.sorting.append(args.index(dist))

    def _cdf(self, x, graph):

        dim,size = x.shape
        out = np.empty((dim,1,size))
        x = x.reshape(dim,1,size)

        for i in self.sorting[::-1]:
            key = "_%03d" % i
            if key in graph.keys:
                out[i] = graph.keys[key]
            else:
                out[i] = graph(x[i], graph.dists[key])

        return out.reshape(dim,size)

    def _mom(self, K, graph):

        if self.dependent():
            raise NotImplementedError()

        dim,size = K.shape
        K = K.reshape(dim,1,size)

        out = np.ones(K.shape[-1])
        for i in self.sorting:
            out *= graph(K[i], graph.dists["_%03d" % i])

        return out

    def _pdf(self, x, graph):

        dim,size = x.shape
        out = np.empty((dim,1,size))
        x = x.reshape(dim,1,size)

        for i in self.sorting[::-1]:
            key = "_%03d" % i
            if key in graph.keys:
                out[i] = graph.keys[key]
            else:
                out[i] = graph(x[i], graph.dists[key])

        return out.reshape(dim,size)

    def _ppf(self, q, graph):

        dim,size = q.shape
        out = np.empty((dim,size))
        q = q.reshape(dim,1,size)

        for i in self.sorting[::-1]:
            key = "_%03d" % i
            if key in graph.keys:
                out[i] = graph.keys[key]
            else:
                out[i] = graph(q[i], graph.dists[key])

        return out.reshape(dim,size)

    def _ttr(self, K, graph):

        if self.dependent():
            raise NotImplementedError("dependency")

        dim,size = K.shape
        K = K.reshape(dim,1,size)

        out = np.zeros((2,dim,size))
        for i in self.sorting:
            out[:, i] = graph(K[i], graph.dists["_%03d" % i])[:,0]

        return out

    def _bnd(self, x, graph):


        dim,size = x.shape
        bnd = np.empty((2,dim,1,size))
        x = x.reshape(dim,1,size)

        for i in self.sorting[::-1]:

            key = "_%03d" % i
            if key in graph.keys:
                bnd[:,i] = graph.keys[key]
            else:
                bnd[:,i] = graph(x[i], graph.dists[key])

        return bnd.reshape(2, dim, size)

    def _val(self, graph):

        if len(graph.keys)!=len(self):
            return self
        out = np.array([graph.keys["_%03d" % i] \
                for i in self.sorting])
        out = out.reshape(len(out), out.shape[-1])
        return out

    def _str(self, **prm):
        dists = [prm["_%03d" % i] \
                for i in range(self.length)]
        dists = ",".join(map(str, dists))
        return "J(" + dists + ")"

    def _dep(self, graph):
        dists = [self.prm["_%03d" % i] \
                for i in range(len(self))]
        sets = [graph(dist)[0] for dist in dists]
        return sets

    def __getitem__(self, i):

        if isinstance(i, int):
            i = "_%03d" % i
            if i in self.prm:
                return self.prm[i]
            raise IndexError()

        if isinstance(i, slice):
            start, stop, step = i.start, i.stop, i.step
            if start is None: start = 0
            if stop is None: stop = len(self)
            if step is None: step = 1
            out = []
            prm = self.prm
            for i in range(start, stop, step):
                out.append(prm["_%03d" % i])
            return J(*out)

        raise NotImplementedError("index not recogniced")


def J(*args):
    """
    Joint random variable generator.

    Args:
        *args : Dist
            Distribution to join together

    Returns:
        Multivariate distribution

    Examples:
        Independent
        >>> dist = chaospy.J(chaospy.Uniform(), chaospy.Normal())
        >>> print(dist.mom([[0,0,1], [0,2,2]]))
        [1.  1.  0.5]

        Dependent
        >>> d0 = chaospy.Uniform()
        >>> dist = chaospy.J(d0, d0+chaospy.Uniform())
        >>> print(dist.mom([[0,0,1], [0,1,1]]))
        [1.  1.  0.5]
    """
    out = []
    args = list(args)
    while args:
        dist = args.pop(0)
        if isinstance(dist, Joint):
            prm = dist.prm
            args = [prm["_%03d" % i] for i in range(len(dist))] + args
        else:
            out.append(dist)

    return Joint(*out)


class Iid(Dist):
    """
    Opaque method for creating independent identical distributed random
    variables from an univariate variable.

    Examples:
        >>> X = chaospy.Normal()
        >>> Y = chaospy.Iid(X, 4)
        >>> chaospy.seed(1000)
        >>> print(Y.sample())
        [ 0.39502989 -1.20032309  1.64760248 -0.04465437]
        """

    def __init__(self, dist, N):
        """
        Parameters:
            dist : Dist
                Input variable. Must have `len(dist)==1`.
            N : int
                Number of variable in the output.
        """
        assert len(dist)==1 and N>1
        Dist.__init__(self, dist=dist, _length=N)

    def _pdf(self, x, dist):
        return dist.pdf(x)

    def _cdf(self, x, dist):
        return dist.fwd(x)

    def _ppf(self, q, dist):
        return dist.inv(q)

    def _bnd(self, dist):
        return dist.range()

    def _mom(self, k, dist):
        return np.prod(dist.mom(k), 0)

    def _ttr(self, k, dist):
        return dist.ttr(k)

    def _str(self, dist):
        return "[%s]%d" % (dist, len(self))

    def __getitem__(self, i):

        if isinstance(i, int):
            if (i >= len(self)) or (i < -len(self)):
                raise IndexError("dist index out of range")
            return self.prm["dist"]

        raise NotImplementedError("index not recogniced")

    def _dep(self, graph):
        dist = graph.dists["dist"]
        return [set([copy(dist)]) for _ in range(len(self))]



import numpy as np

import chaospy
from chaospy.poly.base import Poly, sort_key


def dimsplit(P):
    """
    Segmentize a polynomial (on decomposed form) into it's dimensions.

    In array missing values are padded with 1 to make dimsplit compatible with
    ``poly.prod(Q, 0)``.


    Args:
        P (Poly):
            Input polynomial.

    Returns:
        (Poly):
            Segmentet polynomial array where ``Q.shape==P.shape+(P.dim+1,)``.
            The surplus element in ``P.dim+1`` is used for coefficients.

    Examples:
        >>> x,y = chaospy.variable(2)
        >>> P = chaospy.Poly([2, x*y, 2*x])
        >>> Q = chaospy.dimsplit(P)
        >>> print(Q)
        [[2, 1, 2], [1, q0, q0], [1, q1, 1]]
        >>> print(chaospy.prod(Q, 0))
        [2, q0q1, 2q0]
    """
    P = P.copy()

    if not chaospy.poly.caller.is_decomposed(P):
        raise TypeError("Polynomial not on component form.")
    A = []

    dim = P.dim
    coef = P(*(1,)*dim)
    M = coef!=0
    zero = (0,)*dim
    ones = [1]*dim
    A = [{zero: coef}]

    if zero in P.A:

        del P.A[zero]
        P.keys.remove(zero)

    for key in P.keys:
        P.A[key] = (P.A[key]!=0)

    for i in range(dim):

        A.append({})
        ones[i] = np.nan
        Q = P(*ones)
        ones[i] = 1
        if isinstance(Q, np.ndarray):
            continue
        Q = Q.A

        if zero in Q:
            del Q[zero]

        for key in Q:

            val = Q[key]
            A[-1][key] = val

    A = [Poly(a, dim, None, P.dtype) for a in A]
    P = Poly(A, dim, None, P.dtype)
    P = P + 1*(P(*(1,)*dim)==0)*M

    return P


def setdim(P, dim=None):
    """
    Adjust the dimensions of a polynomial.

    Output the results into Poly object

    Args:
        P (Poly) : Input polynomial
        dim (int) : The dimensions of the output polynomial. If omitted,
                increase polynomial with one dimension. If the new dim is
                smaller then P's dimensions, variables with cut components are
                all cut.

    Examples:
        >>> x,y = chaospy.variable(2)
        >>> P = x*x-x*y
        >>> print(chaospy.setdim(P, 1))
        q0^2
    """
    P = P.copy()

    ldim = P.dim
    if not dim:
        dim = ldim+1

    if dim==ldim:
        return P

    P.dim = dim
    if dim>ldim:

        key = np.zeros(dim, dtype=int)
        for lkey in P.keys:
            key[:ldim] = lkey
            P.A[tuple(key)] = P.A.pop(lkey)

    else:

        key = np.zeros(dim, dtype=int)
        for lkey in P.keys:
            if not sum(lkey[ldim-1:]) or not sum(lkey):
                P.A[lkey[:dim]] = P.A.pop(lkey)
            else:
                del P.A[lkey]

    P.keys = sorted(P.A.keys(), key=sort_key)
    return P

if __name__ == "__main__":
    import doctest
    doctest.testmod()



"""
Implementation of the Golub-Welsh algorithm.
"""
import numpy
import scipy.linalg

import chaospy.quad


def quad_golub_welsch(order, dist, accuracy=100, **kws):
    """
    Golub-Welsch algorithm for creating quadrature nodes and weights.

    Args:
        order (int) : Quadrature order
        dist (Dist) : Distribution nodes and weights are found for with
            `dim=len(dist)`
        accuracy (int) : Accuracy used in discretized Stieltjes procedure. Will
            be increased by one for each itteration.

    Returns:
        (numpy.array, numpy.array) : Optimal collocation nodes with
            `x.shape=(dim, order+1)` and weights with `w.shape=(order+1,)`.

    Examples:
        >>> Z = chaospy.Normal()
        >>> x, w = chaospy.quad_golub_welsch(3, Z)
        >>> print(numpy.around(x, 4))
        [[-2.3344 -0.742   0.742   2.3344]]
        >>> print(numpy.around(w, 4))
        [0.0459 0.4541 0.4541 0.0459]

        Multivariate
        >>> Z = chaospy.J(chaospy.Uniform(), chaospy.Uniform())
        >>> x, w = chaospy.quad_golub_welsch(1, Z)
        >>> print(numpy.around(x, 4))
        [[0.2113 0.2113 0.7887 0.7887]
         [0.2113 0.7887 0.2113 0.7887]]
        >>> print(numpy.around(w, 4))
        [0.25 0.25 0.25 0.25]
    """
    order = numpy.array(order)*numpy.ones(len(dist), dtype=int)+1
    _, _, coeff1, coeff2 = chaospy.quad.generate_stieltjes(
        dist, numpy.max(order), accuracy=accuracy, retall=True, **kws)

    dimensions = len(dist)
    abscisas, weights = _golbub_welsch(order, coeff1, coeff2)

    if dimensions == 1:
        abscisa = numpy.reshape(abscisas, (1, order[0]))
        weight = numpy.reshape(weights, (order[0],))
    else:
        abscisa = chaospy.quad.combine(abscisas).T
        weight = numpy.prod(chaospy.quad.combine(weights), -1)

    assert len(abscisa) == dimensions
    assert len(weight) == len(abscisa.T)
    return abscisa, weight


def _golbub_welsch(orders, coeff1, coeff2):
    """Recurrence coefficients to abscisas and weights."""
    abscisas, weights = [], []

    for dim, order in enumerate(orders):
        if order:
            bands = numpy.empty((2, order))
            bands[0] = coeff1[dim, :order]
            bands[1, :-1] = numpy.sqrt(coeff2[dim, 1:order])
            vals, vecs = scipy.linalg.eig_banded(bands, lower=True)

            abscisa, weight = vals.real, vecs[0, :]**2
            indices = numpy.argsort(abscisa)
            abscisa, weight = abscisa[indices], weight[indices]

        else:
            abscisa, weight = numpy.array([coeff1[dim, 0]]), numpy.array([1.])

        abscisas.append(abscisa)
        weights.append(weight)
    return abscisas, weights

if __name__ == "__main__":
    import doctest
    doctest.testmod()



"""
Multiplication of distributions.

Example usage
-------------

Distribution * a constant::

    >>> distribution = chaospy.Uniform(0, 1) * 4
    >>> print(distribution)
    Uniform(0,1)*4
    >>> print(numpy.around(distribution.sample(5), 4))
    [2.6144 0.46   3.8011 1.9288 3.4899]
    >>> print(numpy.around(distribution.fwd([1, 2, 3]), 4))
    [0.25 0.5  0.75]
    >>> print(numpy.around(distribution.inv(distribution.fwd([1, 2, 3])), 4))
    [1. 2. 3.]
    >>> print(numpy.around(distribution.pdf([1, 2, 3]), 4))
    [0.25 0.25 0.25]
    >>> print(numpy.around(distribution.mom([1, 2, 3]), 4))
    [ 2.      5.3333 16.    ]
    >>> print(numpy.around(distribution.ttr([1, 2, 3]), 4))
    [[2.     2.     2.    ]
     [1.3333 1.0667 1.0286]]

Construct joint addition distribution::

    >>> lhs = chaospy.Uniform(-1, 0)
    >>> rhs = chaospy.Uniform(-3, -2)
    >>> multiplication = lhs * rhs
    >>> print(multiplication)
    Uniform(-1,0)*Uniform(-3,-2)
    >>> joint1 = chaospy.J(lhs, multiplication)
    >>> joint2 = chaospy.J(rhs, multiplication)

Generate random samples::

    >>> print(numpy.around(joint1.sample(4), 4))
    [[-0.7877 -0.9593 -0.6028 -0.7669]
     [ 2.2383  2.1172  1.6532  1.8345]]
    >>> print(numpy.around(joint2.sample(4), 4))
    [[-2.8177 -2.2565 -2.9304 -2.1147]
     [ 2.6843  2.1011  1.2174  0.0613]]

Forward transformations::

    >>> lcorr = numpy.array([-0.9, -0.5, -0.1])
    >>> rcorr = numpy.array([-2.99, -2.5, -2.01])
    >>> print(numpy.around(joint1.fwd([lcorr, lcorr*rcorr]), 4))
    [[0.1  0.5  0.9 ]
     [0.99 0.5  0.01]]
    >>> print(numpy.around(joint2.fwd([rcorr, lcorr*rcorr]), 4))
    [[0.01 0.5  0.99]
     [0.9  0.5  0.1 ]]

Inverse transformations::

    >>> print(numpy.around(joint1.inv(joint1.fwd([lcorr, lcorr*rcorr])), 4))
    [[-0.9   -0.5   -0.1  ]
     [ 2.691  1.25   0.201]]
    >>> print(numpy.around(joint2.inv(joint2.fwd([rcorr, lcorr*rcorr])), 4))
    [[-2.99  -2.5   -2.01 ]
     [ 2.691  1.25   0.201]]

Raw moments::

    >>> print(joint1.mom([(0, 1, 1), (1, 0, 1)]))
    [ 1.25  -0.5   -0.625]
    >>> print(joint2.mom([(0, 1, 1), (1, 0, 1)]))
    [ 1.25  -2.5   -3.125]
"""
import numpy

from ..baseclass import Dist


class Mul(Dist):
    """Multiplication."""

    def __init__(self, left, right):
        """
        Constructor.

        Args:
            left (Dist, array_like) : Left hand side.
            right (Dist, array_like) : Right hand side.
        """
        left_ = not isinstance(left, Dist) or 1 and len(left)
        right_ = not isinstance(right, Dist) or 1 and len(right)
        length = max(left_, right_)
        Dist.__init__(
            self, left=left, right=right, _length=length, _advance=True)


    def _bnd(self, xloc, graph):
        """Distribution bounds."""
        if "left" in graph.keys and "right" in graph.dists:
            num, dist = graph.keys["left"], graph.dists["right"]
        else:
            num, dist = graph.keys["right"], graph.dists["left"]

        num = (num + 1.*(num == 0))
        xloc = (xloc.T/num.T).T
        lower_, upper_ = graph(xloc, dist)
        lower = (num.T*(lower_.T*(num.T > 0) + upper_.T*(num.T < 0))).T
        upper = (num.T*(upper_.T*(num.T > 0) + lower_.T*(num.T < 0))).T

        return lower, upper

    def _cdf(self, xloc, graph):
        """Cumulative distribution function."""
        if "left" in graph.keys and "right" in graph.dists:
            num, dist = graph.keys["left"], graph.dists["right"]
        else:
            num, dist = graph.keys["right"], graph.dists["left"]
        xloc = (xloc.T/(num.T + 1.*(num.T == 0))).T
        out = graph(xloc, dist)
        out = (out.T*(num.T > 0) + (1.-out.T)*(num.T < 0) + \
                1.*(xloc.T > num.T)*(num.T == 0)).T
        return out

    def _ppf(self, uloc, graph):
        """Point percentile function."""
        if "left" in graph.keys and "right" in graph.dists:
            num, dist = graph.keys["left"], graph.dists["right"]
        else:
            num, dist = graph.keys["right"], graph.dists["left"]
        uloc = (uloc*(num > 0) + (1.-uloc)*(num <= 0))
        return graph(uloc, dist)*num

    def _pdf(self, xloc, graph):
        """Probability density function."""
        if "left" in graph.keys and "right" in graph.dists:
            num, dist = graph.keys["left"], graph.dists["right"]
        else:
            num, dist = graph.keys["right"], graph.dists["left"]
        num = numpy.where(num, num, numpy.inf)
        return numpy.abs(graph(xloc*1./num, dist)/num)

    def _mom(self, key, graph):
        """Statistical moments."""
        if len(graph.dists) == 2 and\
                graph.dists["left"].dependent(graph.dists["right"]):
            raise NotImplementedError("dependency")

        left = []
        if "left" in graph.dists:
            left.append(graph(key, graph.dists["left"]))
        else:
            left.append((graph.keys["left"].T**key.T).T)
        if "right" in graph.dists:
            left.append(graph(key, graph.dists["right"]))
        else:
            left.append((graph.keys["right"].T**key.T).T)

        return numpy.prod(left[0]*left[1], 0)

    def _ttr(self, order, graph):
        """Three terms recursion coefficients."""
        if "left" in graph.keys and "right" in graph.dists:
            num = graph.keys["left"]
            dist = graph.dists["right"]
        else:
            dist = graph.dists["left"]
            num = graph.keys["right"]

        coeff0, coeff1 = graph(order, dist)
        return coeff0*num, coeff1*num*num

    def _val(self, graph):
        """Value extraction."""
        if len(graph.keys) == 2:
            return graph.keys["left"]*graph.keys["right"]
        return self

    def _str(self, left, right):
        """String representation."""
        return str(left) + "*" + str(right)


class Mvmul(Dist):
    """Multiplication for multivariate variables."""

    def __init__(self, dist, C):
        """
        Constructor.

        Args:
            dist (Dist, array_like) : Probability.
            C (numpy.ndarray) : matrix to multiply with.
        """
        C = C*numpy.eye(len(dist))
        Dist.__init__(self, dist=dist, C=C,
                Ci=numpy.linalg.inv(C),
                _length=len(dist), _advance=True)

    def _cdf(self, xloc, graph):
        """Cumulative distribution function."""
        return graph(numpy.dot(graph.keys["Ci"], xloc), graph.dists["dist"])

    def _ppf(self, q, graph):
        """Point percentile function."""
        return numpy.dot(graph.keys["C"], graph(q, graph.dists["dist"]))

    def _bnd(self, xloc, graph):
        """Distribution bounds."""
        bnd = graph(xloc, graph.dists["dist"])
        C = graph.keys["C"]
        lower = (numpy.dot(C, bnd[0]).T).T
        upper = (numpy.dot(C, bnd[1]).T).T

        wrong = lower>upper
        out = numpy.where(wrong, upper, lower), numpy.where(wrong, lower, upper)
        return out

    def _val(self, graph):
        """Value extraction."""
        if "dist" in graph.keys:
            return numpy.dot(graph.keys["dist"].T, graph.keys["C"].T).T
        return self

    def _str(self, C, Ci, dist):
        """String representation."""
        return str(dist) + "*" + str(C)

    def _dep(self, graph):
        """Dependency evaluation."""
        dist = graph.dists["dist"]
        S = graph(dist)

        out = [set([]) for _ in range(len(self))]
        C = graph.keys["C"]

        for i in range(len(self)):
            for j in range(len(self)):
                if C[i,j]:
                    out[i].update(S[j])

        return out


def mul(left, right):
    """
    Distribution multiplication.

    Args:
        left (Dist, array_like) : left hand side.
        right (Dist, array_like) : right hand side.
    """
    if left is right:
        return pow(left, 2)

    if isinstance(left, Dist):

        if not isinstance(right, Dist):
            right = numpy.array(right)
            if right.size == 1:
                if right == 1:
                    return left
                if right == 0:
                    return 0.

    elif isinstance(right, Dist):

        left = numpy.array(left)
        if left.size == 1:
            if left == 1:
                return right
            if left == 0:
                return 0.

    else:
        return left*right

    a = not isinstance(left, Dist) or 1 and len(left)
    b = not isinstance(right, Dist) or 1 and len(right)
    length = max(a, b)
    if length == 1:
        return Mul(left, right)
    return Mvmul(dist=left, C=right)



#!/usr/bin/env python
import inspect
import os
import re
from setuptools import setup, find_packages

# move to current directory:
ORIGINAL_PATH = os.path.dirname(
    os.path.abspath(inspect.getfile(inspect.currentframe())))
os.chdir(ORIGINAL_PATH)

with open(os.path.join("src", "chaospy", "version.py")) as src:
    regex = r"^__version__\s*=\s*['\"]([^'\"]+)['\"]"
    VERSION = re.search(regex, src.read(), flags=re.M).group(1)

setup(
    name='chaospy',
    version=VERSION,
    url='https://github.com/jonathf/chaospy',
    author="Jonathan Feinberg",
    author_email="jonathf@gmail.com",
    license='BSD',
    platforms='any',
    packages=find_packages("src"),
    package_dir={"": "src"},
    install_requires=[
        "numpy", "scipy", "networkx"
    ],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Mathematics',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
    ],
)
